from docassemble.base.functions import word


# See GitHub issue https://github.com/SuffolkLITLab/docassemble-ALToolbox/issues/16
def copy_button_html(
    text_to_copy: str,
    text_before: str = "",
    label: str = "Copy",
    tooltip_inert_text: str = "Copy to clipboard",
    tooltip_copied_text: str = "Copied!",
    copy_template_block: bool = False,
    scroll_class: str = "",
    style_class: str = "",
    adjust_height: str = "",
) -> str:
    """Return the html for a button that will let a user copy the given text

    To use, include `docassemble.ALToolbox:copy_button.yml` in your DA interview.

    Args:
      text_to_copy: text you want the user to be able to copy.
      text_before: the prompt that will appear to the left of the HTML input
      label: the label of the Button
      tooltip_inert_text: text shown when the button is hovered over, before it's clicked
      tooltip_copied_text: text shown when the button is hovered over, after the text is placed on the clipboard
    """

    button_str = '<div class="al_copy">\n'
    if text_before != "":
        button_str += (
            f'<span class="al_copy_description">{word( text_before )}</span>\n'
        )

    # Add textarea tag if copy_template_block is True, along with docassemble template block class names
    if copy_template_block:
        button_str += f'<textarea readonly class="card card-body {style_class} bg-light pb-1 al_copy_value {scroll_class}" {adjust_height}>{ text_to_copy }"</textarea>\n'

    # Add input tag if copy_template_block is False
    else:
        button_str += f'<input readonly class="al_copy_value" type="text" value="{ text_to_copy }">\n'

    # Add the copy button
    if copy_template_block:
        button_str += f'<button class="btn btn-secondary al_copy_button al_copy_block" type="button">\n'
    else:
        button_str += (
            f'<button class="btn btn-secondary al_copy_button" type="button">\n'
        )

    # Add tooltip texts
    button_str += f'<span class="al_tooltip al_tooltip_inert">{word( tooltip_inert_text )}</span>\n'
    button_str += f'<span class="al_tooltip al_tooltip_active">{word( tooltip_copied_text )}</span>\n'

    # Add icon and label for the button
    button_str += f'<i class="far fa-copy"></i>\n'
    button_str += f"<span>{word( label )}</span></button>\n"
    button_str += f"</div>\n"

    return button_str



# This function creates a Table list to be used in an addendum file. Currently it handles 'Thing' and 'Inidvidual' object type of DAList (special name attribute).
class myTable:
    def __init__(self, tblData, tblTitle, tblHeader):
        # 1. Put the DAList items into a regular list
        recordList = list()
        for w in tblData:
            my_dict = safe_json2(w)
            for key in list(my_dict):
                # determine the class type
                if key == "_class":
                    if my_dict[key] == "docassemble.base.util.Individual":
                        indicator = "Individual"
                    elif my_dict[key] == "docassemble.base.util.Thing":
                        indicator = "thing"
                    else:
                        indicator = "nothing"  # need refinement
                    my_dict.pop(key)  # remove the class item
                elif (
                    key == "instanceName"
                    or key == "address"
                    or key == "location"
                    or key.startswith("_")
                    or key == "complete"
                ):
                    # remove the item from the dictionary
                    my_dict.pop(key)

                # use the indicator to get rid of extra stuff in the name item
                if key == "name":
                    if indicator == "Individual":
                        my_dict[key] = (
                            my_dict[key]["first"] + " " + my_dict[key]["last"]
                        )
                    elif indicator == "thing":
                        my_dict[key] = my_dict["name"]["text"]
            # Save it to a list
            recordList.append(my_dict)

        # 2. Store the table data in a list for the addendum.
        self.tableList = list()
        self.tableList.append(
            {"tbl_title": tblTitle, "headers": tblHeader, "value": recordList[1:]}
        )


# This function creates a Text Fields list, which can then be used in an addendum file. The limit input is needed to determine if the text field input is too long to fit in the main form.
class myTextList:
    def __init__(self, text, limit, title):
        self.g(text, limit, title)

    def g(self, text, limit, title):
        # 1. Adjust limit
        sLimit = (
            limit - 16
        )  # 16 gives the standard room for inserting " (See Addendum)"
        # 2. Evaluate and shorten a text/area variable
        need_addendum = len(text) > limit
        self.txtList = list()
        if need_addendum:
            self.text_cutoff = text[:sLimit] + " (See Addendum.)"
            # 3. Store the cutoff variable and its lable in txtFieldsList, whitch is referenced in the addendum file.
            self.txtList.append({"text_title": title, "value": text})
        else:
            self.text_cutoff = text


# Function safe_json2 is a revision of Jonathan's function safe_json - mainly to change the date format from string to date.
import types
import datetime
import decimal
import re

TypeType = type(type(None))


def safe_json2(the_object, level=0, is_key=False):
    if level > 20:
        return "None" if is_key else None
    if isinstance(the_object, (str, bool, int, float)):
        return the_object
    if isinstance(the_object, list):
        return [safe_json2(x, level=level + 1) for x in the_object]
    if isinstance(the_object, dict):
        new_dict = dict()
        used_string = False
        used_non_string = False
        for key, value in the_object.items():
            the_key = safe_json2(key, level=level + 1, is_key=True)
            if isinstance(the_key, str):
                used_string = True
            else:
                used_non_string = True
            new_dict[the_key] = safe_json2(value, level=level + 1)
        if used_non_string and used_string:
            corrected_dict = dict()
            for key, value in new_dict.items():
                corrected_dict[str(key)] = value
            return corrected_dict
        return new_dict
    if isinstance(the_object, set):
        new_list = list()
        for sub_object in the_object:
            new_list.append(safe_json2(sub_object, level=level + 1))
        return new_list

    if isinstance(the_object, datetime.datetime):
        # serial = the_object.date()
        serial = the_object.date().strftime("%m/%d/%Y")
        return serial
    if isinstance(the_object, datetime.time):
        serial = the_object.isoformat()
        return serial
    if isinstance(the_object, decimal.Decimal):
        return float(the_object)

    from docassemble.base.core import DAObject

    if isinstance(the_object, DAObject):
        new_dict = dict()
        new_dict["_class"] = type_name(the_object)
        if (
            the_object.__class__.__name__ == "DALazyTemplate"
            or the_object.__class__.__name__ == "DALazyTableTemplate"
        ):
            if hasattr(the_object, "instanceName"):
                new_dict["instanceName"] = the_object.instanceName
            return new_dict
        for key, data in the_object.__dict__.items():
            if key in ["has_nonrandom_instance_name", "attrList"]:
                continue
            new_dict[safe_json2(key, level=level + 1, is_key=True)] = safe_json2(
                data, level=level + 1
            )
        return new_dict
    try:
        json.dumps(the_object)
    except:
        return "None" if is_key else None
    return the_object


def type_name(the_object):
    name = str(type(the_object))
    m = re.search(r"\'(.*)\'", name)
    if m:
        return m.group(1)
    return name



from docassemble.base.util import (
    CustomDataType,
    DAValidationError,
    word,
    as_datetime,
    today,
    date_difference,
    log,
)
from typing import Optional
import re

js_text = """\
// This is an adaptation of Jonathan Pyle's datereplace.js

/*
* Notes to keep around:
* - Rule names must avoid dashes and underscores because of plugin and da limitations.
* - Avoid a min date default for birthdays. Too hard to predict developer needs, like great-grandmother's birthday. Document that developers need to set a min value if they want one.
* - Cannot use regular `min`/`max` attributes. We can get that the rules exist, but we can't get their messages.
* 
* TODO: (post MVP)
* - Get the 'Month'/'Day'/'Year' word values as mako_parameters too so they can be translated. Same as we do within the message functions.
* - Deal with international user date input and international dev date attributes.
* - Handle non-US formatted dates.
* - Consider decreasing padding-right of the day and year fields as sometimes extra numbers get hidden. e.g. A day of 331 only shows the 33. Maybe remove the padding completely to make sure it's clear the text has gone past the edge.
*
* Post MVP validation priority (https://design-system.service.gov.uk/components/date-input/#error-messages):
*  1. missing or incomplete information (when parent is no longer in focus, highlight fields missing info?)
*  2. information that cannot be correct (for example, the number ‘13’ in the month field, 3-digit year)
*  3. information that fails validation for another reason
* Or validate from left to right
*
* Bugs:
* Bad things happen when a user submits invalid fields (upstream?). Note that no other invalidation highlighting appears. If any, it's only our new stuff:
* - Required:
*   - (partly solved by `add_to_groups()` - outline is restored after new on-page error) Any on-page invalid message(s) with any fields filled, message will lose the outline, but only have 1 message. Thereafter, the outline for on-page invalidation will appear, but won't contain the error message.
*   - Required fields with default values don't get invalidated when they are first emptied. They have to have a value put in them and then that value removed. (Upstream?)
* - Sometimes after submission, one validation message will be inside the outline and another outside it. This needs to be replicated consistently to see if `add_to_groups()` can solve this. (upstream?)
* - When one non-required field is empty on submission and there is no on-page invalidation error, the error appears with no red outline. After that, getting a non-submission invalidation message will show the red outline, but the error will be outside of it. Not solved by `add_to_groups()`.
* - Note: At the very least, when submitting before getting a three parts date on-page validation message, our custom errorPlacement doesn't run. There may be other times too. Maybe try moving any existing error on daPageLoad, though not sure this runs after form submission. (Upstream?)
* - Missing error messages: Submit validation only shows an error for one three parts date, even if there are multiple fields on the page. This is a preexisting bug. (Upstream?)
*/

// da doesn't log the full error sometimes, so we'll do our own try/catch
try {{

$(document).on('daPageLoad', function(){{
  $('input[type="ThreePartsDate"]').each(function(){{
    let $al_date = replace_date(this);
    set_up_validation($al_date);
  }});  // ends for each date datatype
}});  // ends on da page load
  
  
function replace_date(original_date) {{
  /** Replace the original date element with our 3 fields and
  *   make sure the fields update the original date value.
  * 
  * @param {{HTML Node}} date The original date element.
  * 
  * @returns {{jQuery obj}} AL container for all the three parts date elements
  */
  let $original_date = $(original_date);
  $original_date.hide();
  $original_date.attr('type', 'hidden');
  $original_date.attr('aria-hidden', 'true');
  
  var $al_date = $('<div class="al_three_parts_date form-row row">');
  $original_date.before($al_date);

  var date_id = $original_date.attr('id');
  
  // -- Construct the input components --
  let $month = create_month(date_id);
  let $day = create_date_part({{date_id, type: 'day'}});
  let $year = create_date_part({{date_id, type: 'year'}});
      
  // -- Add them to the DOM --
  $al_date.append($month.closest('.col'));
  $al_date.append($day.closest('.col'));
  $al_date.append($year.closest('.col'));
  add_error_container($al_date);
  
  // --- Use other original date features ---
  // Avoid .data() for our dynamic stuff - caching problems
  // https://forum.jquery.com/topic/jquery-data-caching-of-data-attributes
  // https://stackoverflow.com/a/8708345/14144258
  
  // Also, original field gets disabled by da on load. Our fields are added
  // after that, so da can't affect them. Must do these attrs ourselves.
  if (is_required($al_date)) {{
    $year.attr('required', true);
    $month.attr('required', true);
    $day.attr('required', true);
  }}
  
  if ($original_date[0].disabled) {{
    $month.attr('disabled', true);
    $day.attr('disabled', true);
    $year.attr('disabled', true);
  }}
  
  use_previous_values({{$original_date, $al_date}});
  
  // Ensure original date field is updated when needed so that
  // submitting the form sends the right data.
  // Updates will be triggered when the user leaves an input field.
  $year.on('change', update);
  $month.on('change', update);
  $day.on('change', update);
  function update() {{
    update_original_date($al_date);
  }};
  
  return $al_date;
}};  // Ends replace_date()
  

function create_date_part({{type, date_id}}) {{
  /** Return one date part with a label and input inside a container.
  *   TODO: Should we use the original dates's `name` instead of `id`?
  *   They've been the same so far. Will they always be?
  * 
  * @param {{str}} type 'year' or 'day'
  * @param {{str}} date_id ID of the original date field
  * 
  * @returns undefined
  */
  var $col = $('<div class="col col-3 col-' + type + '">');
  var id = date_id + '_' + type;
  // '_ignore' prevents the field from being submitted, avoiding a da error
  let name =  '_ignore_' + id;
  
  // For python formatting, need to have {{day}} and {{year}}
  let $label = '';
  if (type === 'day') {{
    $label = $('<label for="' + name + '">{day}</label>');
  }} else {{
    $label = $('<label for="' + name + '">{year}</label>');
  }}
  $col.append($label);
  
  // Reconsider type `number`, `inputmode` ("numeric") not fully supported yet
  // (02/09/2023). When it is, change to `type='text'`. Also, when we _do_ remove
  // `type='number'` also add a check in invalid day check to see if either the
  // year or day is not a number. See warning in that function.
  // Avoid attr `pattern` - voice control will enter invalid input:
  // https://github.com/alphagov/govuk-design-system-backlog/issues/42#issuecomment-775103437
  
  // aria-describedby is ok to have, even when the date-part error is
  // not present or is display: none
  var $field = $('<input class="form-control al_field ' + type + ' ' + date_id + '" type="number" min="1" inputmode="numeric">');
  $field.attr('id', id);
  $field.attr('name', name);
  // There's only one message element, so all fields are described by it.
  // Not sure how the validation plugin associates the original input's id with this.
  $field.attr('aria-describedby', date_id + '-error');
  $col.append($field);
  
  return $field;
}};  // Ends create_date_part()

  
function create_month(date_id) {{
  /** Return one month type date part given the original date node id.
  *   TODO: Should we use the original dates's `name` instead of `id`?
  *   They've been the same so far. Will they always be?
  * 
  * @param {{str}} date_id ID of the original date field
  * 
  * @returns undefined
  */
  var $col = $('<div class="col col-month">');
  
  let id = date_id + '_month';
  // '_ignore' prevents the field from being submitted, avoiding a da error
  let name =  '_ignore_' + id;
  
  let $label = $('<label for="' + name + '">{month}</label>');
  $col.append($label);
  
  // aria-describedby is ok to have, even when the date-part error is
  // not present or is display: none
  // `for` is label of field while `aria-describedby` is supplemental info
  // https://developer.mozilla.org/en-US/docs/Web/CSS/display#display_none
  // https://developer.mozilla.org/en-US/docs/Web/Accessibility/ARIA/Attributes/aria-hidden
  
  // There's only one message element, so all fields are described by it.
  // Not sure how the validation plugin associates the original input's id with this.
  var $field = $('<select class="form-select al_field month ' + date_id + '">');
  $field.attr('id', id);
  $field.attr('name', name);
  // There's only one message element, so all fields are described by it
  $field.attr('aria-describedby', date_id + '-error');
  add_month_options($field);
  
  $col.append($field);
  
  return $field;
}};  // Ends create_month()


function add_month_options(select) {{
  /** Add month values to selection field.
  * 
  * @param {{HTML Node | $ obj}} select A <select> node.
  * 
  * @returns undefined
  */
  let $select = $(select);
  
  // "No month selected" option.
  // TODO: Should this have text for accessibility?
  let $blank_opt = $('<option value=""></option>');
  $select.append( $blank_opt );
  
  // Add every calendar month (based on user's computer's date system? lanugage?)
  for(var month=0; month < 12; month++) {{
    let $opt = $("<option>");
    $opt.val((month + 1 ).toString().padStart(2,0));

    // Convert the month number to a month name for the option text
    var date = new Date(1970, month, 1);
    $opt.text(date.toLocaleString('default', {{ month: 'long' }}));

    $select.append($opt);
  }}  // ends for every month
}};  // Ends add_month_options()
  
  
function use_previous_values({{$original_date, $al_date}}) {{
  /** If $original_date has an existing value, set the date fields values to match.
  *   E.g. If we're returning to a variable that has already been defined.
  * 
  * @param {{$ obj}} $original_date jQuery obj of the original date element.
  * @param {{$ obj}} $al_date jQuery obj of the al parent of our three parts date parts.
  * 
  * @returns undefined
  */
  let date_parts;
  if ( $original_date.val() ) {{
    date_parts = $original_date.val().split( '/' );
    // TODO: Take care of dates with a '-' delimeter?
    date_parts.forEach( function( part, index, date_parts ) {{
      let part_int = parseInt( part );
      if (isNaN(part_int)) {{
        date_parts[ index ] = '';
      }} else {{
        date_parts[ index ] = part_int;
      }}
    }});
  }} else {{
    date_parts = null;
  }}
  
  // TODO: Select "" option if month is empty string?
  if (date_parts && date_parts[0]) {{
    // Ensure 1 becomes "01", etc.
    let month_str = date_parts[0].toString().padStart(2,0);
    let $month = $($al_date.find('select.month')[0]);
    $($month.children('option[value="' + month_str + '"]')).prop('selected', true);
  }}
  
  if (date_parts) {{
    $($al_date.find('input.day')[0]).val(date_parts[1]);
    $($al_date.find('input.year')[0]).val(date_parts[2]);
  }}
}};  // Ends use_previous_values()
  
  
function add_error_container($al_date) {{
  /** Add element that will contain all errors.
  * 
  * @param {{$ obj}} $al_date The al parent of our three parts date parts.
  * 
  * @returns undefined
  */
  let $original_date = get_$original_date($al_date);
  let $error = $('<div id="al_' + $original_date.attr('id') + '_error" class="al_error"></div>');
  $al_date.append($error);
  return $error;
}};  // Ends add_error_container()
  

function update_original_date($al_date) {{
  /** Update value in original date field using the values
  *   of the al three parts date parts.
  * 
  * @param {{$ obj}} $original_date The original date element.
  * @param {{$ obj}} $al_date The al parent of our three parts date parts.
  * 
  * @returns undefined
  */
  var data = get_date_data($al_date);
  
  let US_date = data.month + '/' + data.day + '/' + data.year;
  if (US_date === '//') {{
    US_date = '';
  }}
  
  let val_date = US_date;
  
  // // TODO: If all fields are filled in, we can use the user's locale
  // // settings to create the date Should we avoid this? US person only
  // // visiting the UK, not living there? Regardless, right now its not
  // // possible to work out the complexity of storing and validating
  // // dates of multiple formats. Note that `Date` will fill in the
  // // blanks for any empty string.
  // if (data.month !== '' && data.day !== '' && data.year !== '') {{
  //   // ISO date
  //   val_date = new Date(iso_date).toLocaleDateString(undefined, {{ day: '2-digit', month: '2-digit', year: 'numeric' }})
  // }}
  
  get_$original_date($al_date).val(val_date);
}};  // Ends update_original_date()
  
  
// ==================================================
// ==================================================
// === Get elements and element data ===
// ==================================================
// ==================================================

function is_required(element) {{
  /*** Returns true if date value is required, otherwise returns false.
  * 
  * @param {{Node}} element AL three parts date element. Can be parent of date parts.
  * 
  * @returns {{bool}}
  */
  let $original_date = get_$original_date(element);
  return $original_date.closest('.da-form-group').hasClass('darequired');;
}}  // Ends is_required()
  
  
function get_date_data(element) {{
  /**
  * Given an element that holds a part of the date information,
  * return the full date data as an object.
  * 
  * @param {{HTML Node | $ obj}} element Any al three parts date element, including parent.
  * 
  * @returns {{year: str, month: str, day: str}}
  */
  var year_elem = get_$al_date(element).find('input.year')[0];
  var month_elem = get_$al_date(element).find('select.month')[0];
  var day_elem = get_$al_date(element).find('input.day')[0];
  var date_data = {{
    year: $(year_elem).val(),
    month: $(month_elem).val(),
    day: $(day_elem).val(),
  }};
  return date_data;
}};  // Ends get_date_data()
  
  
function is_birthdate(element) {{
  /** If the element is part of a al birthdate field, returns true, otherwise false.
  * 
  * @param {{HTML Node}} element Any al three parts date element, including al parent.
  * @returns {{bool}}
  */
  // For actual environment:
  let birthdate = get_$al_date(element).parent().find('.daBirthDate')[0];
  if (!birthdate) {{
    // For development:
    birthdate = get_$al_date(element).parent().find('.daALBirthDateTestValidation2')[0];
  }}
  return Boolean(birthdate);
}};  // Ends is_birthdate()

  
function get_$original_date(element) {{
  /** Returns jQuery obj of original date element or an empty jQuery object.
  * 
  * @param {{HTML Node}} element Any al three parts date element, including al parent.
  * 
  * @returns {{jQuery obj}}
  */
  return $($(element).closest('.dafieldpart').children('input')[0]);
}};  // Ends get_$original_date()
  
  
function get_$al_date(element) {{
  /** Using any element inside the element containing our three parts date,
  *   return the element we created to surround our date elements.
  *   If it doesn't exist, will return an empty jQuery object.
  *   Easier to maintain all in one place. Consider returning
  *   a plain element - calling functs won't have to know how
  *   to check for an empty jQuery object.
  * 
  * @param {{HTML Node}} element Any element.
  * 
  * @returns {{jQuery obj}} Note: can be an "empty" jQuery object.
  */
  // `.closest()` will get the element itself if appropriate
  return $(element).closest('.dafieldpart').find('.al_three_parts_date');
}};  // Ends get_$al_date()
  
  
// ==================================================
// ==================================================
// === Validation ===
// ==================================================
// ==================================================
  
function set_up_validation($al_date) {{
  /** Uses jQuery validation plugin to set up validation functionality
  * 
  * @param {{$ obj}} $al_date jQuery obj of the al parent of our three parts date parts.
  * 
  * @returns undefined
  */
  $al_date.find('.al_field').each(function make_validator_options (index, field) {{
    add_rules(field);
    add_messages(field);
    add_to_groups(field);
  }});
  
  // There's some strange behavior on submission outlined in the
  // description of bugs. This helps with those involving multiple
  // messages cropping up. Avoid adding rules and messages to the
  // original field for now, though. That seemed to cause issues.
  let $original_date = get_$original_date($al_date);
  add_to_groups($original_date);
  
  let validator = $("#daform").data('validator');
  set_up_errorPlacement(validator);
  set_up_highlight(validator);
  set_up_unhighlight(validator);
}};  // Ends set_up_validation()
  

// ==================================================
// ===== Valdiation plugin configs =====
// ==================================================
  
function add_rules(field) {{
  /** Add all date rules to a given field.
  * 
  * @param {{HTML Node}} field An al three parts date field.
  * 
  * @returns undefined
  */
  let rules = {{
    alMin: get_$original_date(field).attr('data-alMin') || false,
    // TODO: try:
    alMax: is_birthdate(field) || get_$original_date(field).attr('data-alMax'),
    _alInvalidDay: true,  // e.g. 1/54/2000 is invalid. TODO: Should devs be able to disable this?
    _alInvalidYear: true,  // e.g. 200 or 012. TODO: Should devs be able to disable this?
    // Normal `required` only deals with one field being empty, not empty siblings
    _alRequired: is_required(field),
  }};  // ends rules
  
  $(field).rules('add', rules);
}};  // Ends add_rules()
  
  
function add_messages(field) {{
  /** Adds custom messages that are in the validator object
  *   and don't need parameters.
  * 
  * @param {{HTML Node}} field An al three parts date field.
  * 
  * @returns undefined
  */
  let messages = $("#daform").data('validator').settings.messages;
  let name = get_$original_date(field).attr('name');
  
  // If there's are custom messages in the validator object, use them
  if (messages[name]) {{
    $(field).rules('add', {{
      messages: {{
        // We can access the original da `required` message
        required: messages[get_$original_date(field).attr('name')].required,
        // Normal `required` only deals with one field being empty, not empty siblings
        _alRequired: messages[get_$original_date(field).attr('name')].required,
      }}
    }});  // ends add rules
  }}  // ends if required message exists
  
}};  // Ends add_messages()
  
  
function add_to_groups(field) {{
  /** Add field to its group dynamically after-the-fact. We have
  *   to do this because da has already created its groups and we
  *   don't want to override anything.
  *   Note: Adding groups dynamically here won't be reflected in `validator.settings`
  * 
  *   Inspired by https://stackoverflow.com/a/9688284/14144258
  * 
  * @param {{HTML Node}} field An al three parts date field.
  * 
  * @returns undefined
  */
  let groups = $("#daform").data('validator').groups;
  groups[ $(field).attr('name') ] = get_$original_date(field).attr('id');
}};  // Ends add_to_groups()
  
  
// ==================================================
// ===== Validation methods =====
// ==================================================
  
// -- Whole date validations --

$.validator.addMethod('alMin', function(value, field, params) {{
  /** Returns true if full date is >= min date or incomplete. */
  if (!date_is_ready_for_min_max(field)) {{
    return true;
  }}
  
  // Avoid usinig `params`, which could be in many different formats.
  // Just get date data from the actual fields
  var data = get_date_data(field);
  // TODO: Catch invalid `data-alMin` attr values? Useful for devs.
  // Otherwise very hard for devs to track down. Log in console?
  // Non-MVP. Make an issue.
  var date_min = new Date( get_$original_date(field).attr('data-alMin') );
  var date_val = new Date(data.year + '-' + data.month + '-' + data.day);
  
  return date_val >= date_min;
  
}}, function alMinMessage (validity, field) {{
  /** Returns the string of the invalidation message. */
  let min_date = get_$original_date(field).attr('data-alMin');
  let locale_long_date = new Date(min_date).toLocaleDateString(undefined, {{ day: '2-digit', month: 'long', year: 'numeric' }})
  return (
    get_$original_date(field).attr('data-alMinMessage')
    || get_$original_date(field).attr('data-alDefaultMessage')
    // || `The date needs to be on or after ${{ get_$original_date(field).attr('data-alMin') }}.`
    || `The date needs to be on or after ${{ locale_long_date }}.`
  );
}});  // ends validate 'alMin'
  
  
$.validator.addMethod('alMax', function(value, field, params) {{
  /** Returns true if full date is <= max date or is incomplete. */
  if (!date_is_ready_for_min_max(field)) {{
    return true;
  }}

  // Avoid usinig `params`, which could be in many different formats.
  // Just get date data from the actual fields
  var data = get_date_data(field);

  // TODO: Catch invalid alMax attr values for devs? Log in console? Make post MVP issue
  var max_attr = get_$original_date(field).attr('data-alMax');
  var date_max = new Date(max_attr);
  if ( isNaN(date_max) && is_birthdate(field)) {{
    date_max = new Date(Date.now());
  }}
  
  var date_val = new Date(data.year + '-' + data.month + '-' + data.day);
  
  return date_val <= date_max;
  
}}, function alMaxMessage (validity, field) {{
  /** Returns the string of the invalidation message. */
  
  let max_date = get_$original_date(field).attr('data-alMax');
  let locale_long_date = new Date(max_date).toLocaleDateString(undefined, {{ day: '2-digit', month: 'long', year: 'numeric' }})
  var default_MaxMessage = `The date needs to be on or before ${{ locale_long_date }}.`;
  // Birthdays have a different default max message
  if (!get_$original_date(field).attr('data-alMax') && is_birthdate(field)) {{
    default_MaxMessage = 'A <strong>birthdate</strong> must be in the past.';
  }}
  
  return (
    get_$original_date(field).attr('data-alMaxMessage')
    || get_$original_date(field).attr('data-alDefaultMessage')
    || default_MaxMessage
  );
}});  // ends validate 'alMax'
  
  
// --- Validate individual fields ---
// Always check each others' field any time any field is validated. That
// way one valid field can't take away the message/highlighting for another
// invalid field.
  
$.validator.addMethod('_alInvalidDay', function(value, field, params) {{
  /** Returns false if full input values cannot be converted to a
  *   matching Date object. E.g. HTML won't recognize 12/32/2000 as an
  *   invalid date. It will instead convert it to 1/1/2001.
  *   Only day inputs can create mismatching dates, but this must be
  *   checked whenever any field is updated. If the user puts Jan 30 and
  *   then change month to Feb, we need to show the error then. If the
  *   user puts 2/29/2020 and changes to 2021, that must be shown too.
  */
  // Doesn't need to be abstracted anymore in some ways, but it does
  // make this addMethod section of the code cleaner.
  return has_valid_day(field);
}}, function alInvalidDayMessage (validity, field) {{
  /** Returns the string of the invalidation message. */
  
  // Always return a custom message first
  let custom_msg = get_$original_date(field).attr('data-alInvalidDayMessage')
                   || get_$original_date(field).attr('data-alDefaultMessage');
  if (custom_msg) {{
    return custom_msg;
  }}
  
  let input_date = get_$al_date(field).find('.day').val();
  
  // If the date is only partly filled, we can't give a useful message
  // without a heck of  a lot of work, so give a generalized invalid day
  // default message.
  let data = get_date_data(field);
  if (data.year == '' || data.month == '') {{
    return `No month has ${{input_date}} days.`;
  }}
  
  // Otherwise we can give the full default message
  let input_year = get_$al_date(field).find('.year').val();
  let converted_year = (new Date(`1/1/${{input_year}}`)).getFullYear();
  let input_month = get_$al_date(field).find('.month option:selected').text();
  
  return `${{input_month}} ${{converted_year}} doesn't have ${{input_date}} days.`;
}});  // ends validate '_alInvalidDay'
  

$.validator.addMethod('_alInvalidYear', function(value, field, params) {{
  /** Returns true if year is empty or has 4 digits.
  * 
  * @returns {{bool}}
  */
  let text = get_$al_date(field).find('input.year')[0].value;
  // Empty year is not invalid in this way
  if (text.length === 0) {{return true;}}
  // Dates will remove leading 0's, thus sadly 0011 == 2011
  if (text.length !== 4 || text[0] === '0') {{
    return false;
  }} else {{
    return true;
  }}
  
}}, function alInvalidYearMessage (validity, field) {{
  /** Returns the string of the invalidation message. */
  return (
    get_$original_date(field).attr('data-alInvalidYearMessage')
    || get_$original_date(field).attr('data-alDefaultMessage')
    || `The year needs to be 4 digits long and cannot start with "0".`
  );
}});  // ends validate '_alInvalidYear'
  

$.validator.addMethod('_alRequired', function(value, field, params) {{
  /** Returns true if
  *   - original field is hidden/disabled
  *   - all fields have contents
  *   - current field has contents and other fields haven't been interacted with yet
  *   Otherwise returns false.
  *
  * Unable to give a specific message about which date field is missing values
  *   (like in the CustomDataType 'submit' validation messages do) as we'd have
  *   to show a lot of premature error messages.
  * 
  * @returns {{bool}}
  */
  // Remember that this field has been interacted with for validation.
  $(field).addClass('al_dirty');
  
  // For clarity, if the field itself has just been made empty, easy choice
  if ($(field).val() === '') {{
    return false;
  }}
  
  let all_dirty_fields_have_contents = true;
  // For all related three parts date fields
  get_$al_date(field).find('.al_field').each(function (index, a_field) {{
    // If a field has been interacted with by this rule at least once
    if ( $(a_field).hasClass('al_dirty') ) {{
      // and it's now empty
      if ($(a_field).val() === '') {{
        // all fields should be invalid
        all_dirty_fields_have_contents = false;
      }}
    }}
  }});
  
  return all_dirty_fields_have_contents;
}});  // ends validate 'required'
  
  
// ==================================================
// ===== Validation calculations =====
// ==================================================

function date_is_ready_for_min_max(element) {{
  /** Return true if date input is ready to be evaluated for min/max
  *   date value invalidation.
  * 
  * @param {{HTML Node}} element Any al three parts date element, including the parent.
  * 
  * @returns {{bool}}
  */
  var data = get_date_data(element);
  // Don't evaluate min/max if the date is only partly filled
  if (data.year == '' || data.month == '' || data.day === '') {{
    return false;
  }}
  // Not sure how we'd get here, but don't evaluate min/max if the date is
  // invalid in some other way. Maybe negative numbers?
  var date_val = new Date(data.year + '-' + data.month + '-' + data.day);
  if (isNaN(date_val)) {{
    return false;
  }}
  
  // Don't need to avoid an invalid day situation. Seems to work
  // without that check in here.
  
  return true;
}};  // Ends date_is_not_ready_for_min_max()
  

function has_valid_day(element) {{
  /** Given a date part element, returns true if:
  *   - the day is <= 31 and year or month is empty
  *   - all fields are filled and the day date is <= max days in the given month in the given year
  *   Returns false if either:
  *   - The day date is > 31 or
  *   - The day date is > the max days in the given month of the given year
  *
  * Inspired by https://github.com/uswds/uswds/blob/728ba785f0c186e231a81865b0d347f38e091f96/packages/usa-date-picker/src/index.js#L735
  * 
  * @param element {{HTML Node}} Any element in the al three parts date picker
  * 
  * @returns bool True if day date is valid
  * 
  * @examples:
  * 10//2000  // true
  * /10/2000  // true
  * 10/10/2000  // true
  * /42/2000  // false
  * 2/29/2021  // false
  */
  var data = get_date_data(element);

  if (parseInt(data.day) > 31) {{
    return false;
  }}
  // Don't invalidate if the date is only partly filled. Empty input fields
  // should get handled other places.
  if (data.year === '' || data.month === '' || data.day === '') {{
    return true;
  }}

  const dateStringParts = [data.year, data.month, data.day];
  const [year, month, day] = dateStringParts.map((str) => {{
    let value;
    const parsed = parseInt(str, 10);
    if (!Number.isNaN(parsed)) value = parsed;
    return value;
  }});
  
  // WARNING: If we change the `type` of the year or day
  // to be 'text', we need to check return false if any
  // dateStringParts part === null. See inspiration link.
  
  const checkDate = setDate({{
    year: year,
    month: month - 1,
    date: day
  }});

  // 12/32/2000 would have transformed into 1/1/2001
  if (
    checkDate.getFullYear() !== year ||
    checkDate.getMonth() !== (month - 1) ||
    checkDate.getDate() !== day
  ) {{
    return false;
  }}

  return true;
}};  // Ends has_valid_day()

function setDate({{year, month, date}}) {{
  /**
  * Set date from month day year
  *
  * @param {{number}} year the year to set
  * @param {{number}} month the month to set (zero-indexed)
  * @param {{number}} date the date to set
  * 
  * @returns {{Date}} the set date
  */
  const newDate = new Date(0);
  newDate.setFullYear(year, month, date);
  return newDate;
}};
  
  
// ==================================================
// ===== Visual feedback management =====
// ==================================================
  
function set_up_errorPlacement(validator) {{
  /** Sometimes override existing errorPlacement.
  *
  * @param {{obj}} validator The form's validator object.
  * 
  * @returns undefined
  */
  let original_error_placement = validator.settings.errorPlacement;
  validator.settings.errorPlacement = function al_errorPlacement(error, field) {{
    /** Put all errors in one spot at the bottom of the parent.
    *   Only runs once per field.
    *   WARNING: If submission validation happens first, this function will
    *   not be run and the error won't be placed inside our AL date.
    */
    let $al_date = get_$al_date(field);
    // If this isn't an AL date, use the original behavior
    if (!$al_date[0] && original_error_placement !== undefined) {{
      original_error_placement(error, field);
      return;
    }}

    $(error).appendTo($al_date.find('.al_error')[0]);
  }};  // Ends al_errorPlacement()
  
}};  // Ends set_up_errorPlacement()
  
function set_up_highlight(validator) {{
  /** For our date elements, override pre-existing highlight method.
  *
  * @param {{obj}} validator The form's validator object.
  * 
  * @returns undefined
  */
  let original_highlight = validator.settings.highlight;
  validator.settings.highlight = function al_highlight(field, errorClass, validClass) {{
    /** Highlight parent instead of individual fields. MVP */
    let $al_date = get_$al_date(field);
    // If this isn't an AL date, use the original behavior
    if (!$al_date[0] && original_highlight !== undefined) {{
      original_highlight(field, errorClass, validClass);
      return;
    }}
    
    $al_date.addClass('al_invalid');
    // Avoid highlighting individual elements
    $al_date.find('.al_field').each(function(index, field) {{
      $(field).removeClass('is-invalid');  // Just a Bootstrap class
      // TODO: try just the below alone
      $(field).removeClass(errorClass);  // Just a Bootstrap class
    }});
    
  }};  // Ends al_highlight()
}};  // Ends set_up_highlight()
  
function set_up_unhighlight(validator) {{
  /** For our date elements, override pre-existing highlight method.
  *
  * @param {{obj}} validator The form's validator object.
  * 
  * @returns undefined
  */
  let original_unhighlight = validator.settings.unhighlight;
  validator.settings.unhighlight = function al_unhighlight(field, errorClass, validClass) {{
    /** Unhighlight parent instead of individual fields. MVP */
    // During invalid required day, this is triggered for month and unhighlights all. Why?
    let $al_date = get_$al_date(field);
    $al_date.removeClass('al_invalid');
    original_unhighlight(field, errorClass, validClass);
  }};  // Ends al_unhighlight()
}};  // Ends set_up_unhighlight()
  
  
}} catch (error) {{
  console.error('Error in AL three parts date CusotmDataTypes', error);
}}

"""


def check_empty_parts(item: str, default_msg="{} is not a valid date") -> Optional[str]:
    # This only handles US dates. How do we use a locale-specific date?
    parts = item.split("/")
    empty_parts = [part == "" for part in parts]
    if len(empty_parts) != 3:
        return word(default_msg).format(item)
    if not any(empty_parts):
        return None
    if all(empty_parts):
        return word("Enter a month, a day, and a year")
    elif sum(empty_parts) == 2:
        # only on part was given, enumerate each
        if not empty_parts[0]:
            return word("Enter a day and a year")
        elif not empty_parts[1]:
            return word("Enter a month and a year")
        else:
            return word("Enter a month and a day")
    elif sum(empty_parts) == 1:
        if empty_parts[0]:
            return word("Enter a month")
        elif empty_parts[1]:
            return word("Enter a day")
        else:
            return word("Enter a year")
    else:
        return word(default_msg).format(item)


class ThreePartsDate(CustomDataType):
    name = "ThreePartsDate"
    input_type = "ThreePartsDate"
    javascript = js_text.format(month=word("Month"), day=word("Day"), year=word("Year"))
    jq_message = word("Answer with a valid date")
    is_object = True
    # Unable to get messages for plain `min`/`max` attributes
    # Unable to use `-` in names for validation plugin.
    # Unable to use `_` in names which da turns into `-`
    mako_parameters = [
        "alMin",
        "alMinMessage",
        "alMax",
        "alMaxMessage",
        "alInvalidDayMessage",
        "alInvalidYearMessage",
        "alDefaultMessage",
    ]

    @classmethod
    def validate(cls, item: str):
        # If there's no input in the item, it's valid
        if not isinstance(item, str) or item == "":
            return True
        else:
            # Otherwise it needs to be a date after the year 1000. We ourselves make
            # sure this format is created if the user gives valid info.
            matches_date_pattern = re.search(r"^\d{1,2}\/\d{1,2}\/\d{4}$", item)
            if matches_date_pattern:
                try:
                    date = as_datetime(item)
                except Exception as error:
                    ex_msg = f"{ item } {word('is not a valid date')}"
                    raise DAValidationError(ex_msg)
                return True
            else:
                msg = check_empty_parts(item)
                if msg:
                    raise DAValidationError(msg)

    @classmethod
    def transform(cls, item):
        if item:
            return as_datetime(item)

    @classmethod
    def default_for(cls, item):
        if item:
            return item.format("MM/dd/yyyy")


class BirthDate(ThreePartsDate):
    name = "BirthDate"
    input_type = "BirthDate"
    javascript = js_text.format(
        month=word("Month"), day=word("Day"), year=word("Year")
    ).replace("ThreePartsDate", "BirthDate")
    jq_message = word("Answer with a valid date of birth")
    is_object = True
    # Unable to get messages for plain `min`/`max` attributes
    # Unable to use `-` in names for validation plugin.
    # Unable to use `_` in names which da turns into `-`
    mako_parameters = [
        "alMin",
        "alMinMessage",
        "alMax",
        "alMaxMessage",
        "alInvalidDayMessage",
        "alInvalidYearMessage",
        "alDefaultMessage",
    ]

    @classmethod
    def validate(cls, item: str):
        # If there's no input in the item, it's valid
        if not isinstance(item, str) or item == "":
            return True
        else:
            # Otherwise it needs to be a date on or before today and after the year 1000.
            # We ourselves create this format if the user gives valid info.
            matches_date_pattern = re.search(r"^\d{1,2}\/\d{1,2}\/\d{4}$", item)
            try:
                date = as_datetime(item)
            except Exception as error:
                raise DAValidationError(word("{} is not a valid date").format(item))
            if matches_date_pattern:
                date_diff = date_difference(starting=date, ending=today())
                if date_diff.days >= 0.0:
                    return True
                else:
                    raise DAValidationError(
                        word(
                            "Answer with a <strong>date of birth</strong> ({} is in the future)"
                        ).format(item)
                    )
            else:
                msg = check_empty_parts(
                    item, default_msg="{} is not a valid <strong>date of birth</strong>"
                )
                if msg:
                    raise DAValidationError(msg)



# Add up numeric field values from a DAList object.
from docassemble.base.util import DAValidationError, word


class Addup:
    def __init__(self, listName, varName):
        self.g(listName, varName)

    def g(self, listName, varName):
        self.sum = 0
        for w in listName:
            my_dict = w.as_serializable()
            for key, val in my_dict.items():
                if key == varName:
                    self.sum += val
        if self.sum == 0:
            msg = (
                "Make sure your field '"
                + varName
                + "' is spelled correctly and is numeric."
            )
            raise DAValidationError(msg)
        else:
            return self.sum



from docassemble.webapp.jsonstore import (
    read_answer_json,
    write_answer_json,
    JsonStorage,
    JsonDb,
)
from docassemble.base.generate_key import random_alphanumeric
from docassemble.base.functions import get_current_info
from docassemble.base.util import variables_snapshot_connection, user_info, DADict
import random
from typing import Dict, List, Any, Optional

__all__ = ["save_input_data"]


def save_input_data(
    title: str = "",
    input_dict: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
) -> None:
    """
    This function is used by survey type interviews to save input data for data reporting purposes.

    The input_dict should a dictionary where each key is a string and each value is a value from a Docassemble interview
    question. Typically that is a string, float, int, or a DADict.
    """
    type_dict: Dict[str, str] = {}
    field_dict: Dict[str, Any] = {}
    if not input_dict:
        # Can still save the tags, so just use an empty input dict
        input_dict = {}
    for k, v in input_dict.items():
        field_dict[k] = v
        if isinstance(v, int):
            type_dict[k] = "int"
        elif isinstance(v, float):
            type_dict[k] = "float"
        elif isinstance(v, DADict):  # This covers checkboxes and multiselect
            type_dict[k] = "checkboxes"
        else:
            type_dict[k] = "text"

    data_to_save: Dict[str, Any] = {}
    data_to_save["title"] = title

    # TODO(qs): We should be able to infer type in the InterviewStats package too, eventually. But
    # leaving as-is for now
    data_to_save["field_type_list"] = type_dict  # This may not be needed

    for k, v in type_dict.items():
        # If a field is of checkboxes type, flatten its elements dict
        # so that each key/value pair is saved in its own column.
        if v in ["checkboxes", "multiselect"]:
            for subkey, subvalue in field_dict[k].elements.items():
                data_to_save[k + "_" + subkey] = subvalue
        else:
            data_to_save[k] = field_dict[k]

    # Save one record per session to JsonStorage datatable.
    filename = get_current_info().get("yaml_filename", None)
    random_uid = random_alphanumeric(32)
    new_entry = JsonStorage(
        filename=filename,
        key=random_uid,
        data=data_to_save,
        tags=tags,
        persistent=False,
    )
    JsonDb.add(new_entry)
    JsonDb.commit()



from typing import Optional, Union

from base64 import b64encode
from decimal import Decimal
import docassemble.base.functions
from docassemble.base.util import (
    defined,
    value,
    showifdef,
    space_to_underscore,
    action_button_html,
    Address,
    word,
)
import re

__all__ = [
    "shortenMe",
    "thousands",
    "tel",
    "fa_icon",
    "space",
    "yes_no_unknown",
    "number_to_letter",
    "collapse_template",
    "tabbed_templates_html",
    "review_widget",
    "sum_if_defined",
    "add_records",
    "output_checkbox",
    "nice_county_name",
]


class shortenMe:
    def __init__(self, originalURL):
        self.shortenedURL = docassemble.base.functions.temp_redirect(
            originalURL, 60 * 60 * 24 * 7, False, False
        )


# The following three functions are from Quinten
def thousands(num: Union[float, str, Decimal], show_decimals=False) -> str:
    """
    Return a whole number formatted with thousands separator.
    Optionally, format with 2 decimal points (for a PDF form with the
    currency symbol already present in the form)

    If `show_decimals`, will call `int(num)`, truncating the decimals instead of
    rounding to the closest int.
    """
    try:
        if show_decimals:
            return f"{num:,.2f}"
        else:
            return f"{int(num):,}"
    except:
        return str(num)


def tel(phone_number) -> str:
    """Format a phone number so you can click on it to open in your phone dialer"""
    return '<a href="tel:' + str(phone_number) + '">' + str(phone_number) + "</a>"


def fa_icon(
    icon: str,
    color: Optional[str] = "primary",
    color_css: Optional[str] = None,
    size: Optional[str] = "sm",
    fa_class: str = "fa-solid",
) -> str:
    """Display a fontawesome icon inline.

    Docassemble allows you to display an icon from [fontawesome](https://fontawesome.com),
    but it does not provide control over the size or color of the icon. This function gives
    you more control over the icon that is inserted.

    Args:
      icon: a string representing a fontawesome icon. The icon needs to be in the
        [free library](https://fontawesome.com/search?o=r&m=free).
      color: can be any [Bootstrap color variable](https://getbootstrapc.mo/docs/4.0/utilities/colors).
        For example: `primary`, `secondary`, `warning`
      color_css: allows you to use a CSS code to represent the color, e.g., `blue`, or `#fff` for black
      size: used to control the [fontawesome size](https://fontawesome.com/v6.0/docs/web/style/size)
        (without the `fa-` prefix). Valid values include `2xs`, `xs`, the default of `sm`,
        `md`, `lg`, `xl`, `2x1`, and the python `None`, which defaults to `md`.
      fa_class: let's you specify the fontawesome class, needed for any icon that isn't
        the default class of `fa-solid`, like `fa-brands`, or `fa-regular` and `fa-light`.

    Returns:
      HTML for a font-awesome icon of the specified size and color.
    """
    if not size or size == "md":
        size_str = ""
    else:
        size_str = " fa-" + size
    if not size and not color and not color_css:
        return ":" + icon + ":"  # Default to letting Docassemble handle it
    if color_css:
        return (
            f'<i class="{fa_class} fa-'
            + icon
            + size_str
            + '" style="color:'
            + color_css
            + ';"></i>'
        )
    if color:
        return (
            f'<i class="{fa_class} fa-'
            + icon
            + size_str
            + '" style="color:var(--'
            + color
            + ');"></i>'
        )
    return f'<i class="{fa_class} fa-{icon}{size_str}"></i>'


def space(var_name: str, prefix=" ", suffix="") -> str:
    """If the value as a string is defined, return it prefixed/suffixed. Defaults to prefix
    of a space. Helps build a sentence with less cruft. Equivalent to SPACE function in
    HotDocs."""
    if (
        var_name
        and isinstance(var_name, str)
        and re.search(r"[A-Za-z][A-Za-z0-9\_]*", var_name)
        and defined(var_name)
        and value(var_name)
    ):
        return prefix + showifdef(var_name) + suffix
    else:
        return ""


def yes_no_unknown(
    var_name: str, condition: Optional[bool], unknown="Unknown", placeholder=0
):
    """Return 'unknown' if the value is None rather than False. Helper for PDF filling with
    yesnomaybe fields"""
    if condition:
        return value(var_name)
    elif condition is None:
        return unknown
    else:
        return placeholder


def number_to_letter(n: Optional[int]) -> str:
    """Returns a capital letter representing ordinal position. E.g., 1=A, 2=B, etc. Appends letters
    once you reach 26 in a way compatible with Excel/Google Sheets column naming conventions. 27=AA, 28=AB...
    """
    string = ""
    if n is None:
        n = 0
    while n > 0:
        n, remainder = divmod(n - 1, 26)
        string = chr(65 + remainder) + string
    return string


def collapse_template(
    template,
    classname=None,
    closed_icon="caret-right",
    open_icon="caret-down",
):
    """
    Insert HTML for a Bootstrap "collapse" div.

    Optionally, you can specify a custom icon to override the defaults:

    The default icons are "right caret" which displays when the text is collapsed (`closed_icon`) and
    "down caret" which displays when the text is open (`open_icon`).
    """
    if not template.subject_as_html(trim=True) and not template.content_as_html():
        return ""

    if classname is None:
        classname = " bg-light"
    else:
        classname = " " + classname.strip()
    container_classnames = "al_collapse_template"

    container_id = (
        b64encode(str(template.instanceName).encode()).decode().replace("=", "")
    )
    contents_id = f"{ container_id }_contents"

    return f"""\
<div id="{ container_id }" class="{ container_classnames }">
<a class="collapsed al_toggle" data-bs-toggle="collapse" href="#{ contents_id }" role="button" aria-expanded="false" aria-controls="{ contents_id }"><span class="toggle-icon pdcaretopen">{ fa_icon(open_icon) }</span><span class="toggle-icon pdcaretclosed">{ fa_icon(closed_icon) }</span>
<span class="subject">{ template.subject_as_html(trim=True) }</span></a>
<div class="collapse" id="{ contents_id }">
<div class="card card-body pb-1{ classname }">{ template.content_as_html() }</div>
</div>
</div>\
"""


def tabbed_templates_html(tab_group_name: str, *pargs) -> str:
    """
    Provided a list of templates, create Bootstrap v 4.5 tabs with the `subject` as the tab label.
    """
    if isinstance(tab_group_name, str):
        tab_group_name = space_to_underscore(tab_group_name)
    else:
        tab_group_name = "tabbed-group"
    tabs = f'<ul class="nav nav-tabs" id="{tab_group_name}" role="tablist">\n'
    tab_content = '<div class="tab-content" id="myTabContent">'
    for index, templ in enumerate(pargs):
        tab_id = f"{tab_group_name}-{space_to_underscore(str(templ.subject))}"
        tabs += '<li class="nav-item" role="presentation">\n'
        if index == 0:
            tabs += f'<a class="nav-link active" id="{tab_id}-tab" data-bs-toggle="tab" href="#{tab_id}" role="tab" aria-controls="{tab_id}" aria-selected="true">{templ.subject}</a>\n'
            tab_content += f'<div class="tab-pane fade show active" id="{tab_id}" role="tabpanel" aria-labelledby="{tab_id}-tab">\n'
            tab_content += templ.content_as_html()

            tab_content += "\n</div>\n"
        else:
            tabs += f'<a class="nav-link" id="{tab_id}-tab" data-bs-toggle="tab" href="#{tab_id}" role="tab" aria-controls="{tab_id}" aria-selected="false">{templ.subject}</a>\n'
            tab_content += f'<div class="tab-pane fade" id="{tab_id}" role="tabpanel" aria-labelledby="{tab_id}-tab">\n'
            tab_content += templ.content_as_html()
            tab_content += "\n</div>\n"
        tabs += "</li>"
    tabs += "</ul>"
    tab_content += "</div>"

    return tabs + tab_content


def review_widget(
    *,
    up_action: str,
    down_action: str,
    review_action: Optional[str] = None,
    thumbs_display: str = "Did we help you?",
    review_display: str = "Thank you for your feedback. Let us know what we could do better",
    submit_review_button: str = "Add your review",
    post_review_display: str = "Thank you for your review!",
) -> str:
    """
    A widget that allows people to give a quick review (thumbs up and down, with an optional text
    component) in the middle of an interview without triggering a page reload.

    If `review_action` is provided, once you press either of the thumbs, a text input screen with
    a submit button appears, and once the text review is submitted (or after the thumbs, if no
    `review_action` was provided), a final "thank you" message is displayed.

    Args:
        up_action: the variable name of an event to be executed on the server if the
            thumbs up is pressed
        down_action: the variable name of an event to be executed on the server if the
            thumbs down is pressed
        review_action: the variable name of an event to be execute on the
            server when someone submits their text review
        thumbs_display: text displayed to user describing the thumbs
        review_display: text displayed to user describing the text input
        submit_review_button: text on the button to submit their text review
        post_review_display: text displayed to user after review is submitted
    Returns:
        the HTML string of the widget
    """
    js_thumbs_up = f"javascript:altoolbox_thumbs_up_send('{up_action}', {'true' if review_action else 'false'})"
    js_thumbs_down = f"javascript:altoolbox_thumbs_down_send('{down_action}', {'true' if review_action else 'false'})"
    widget = f"""
<div class="card" style="width: 20rem;">
  <div class="card-body">
    <p class="al-thumbs-widget">{word(thumbs_display)}</p>
    <a href="{js_thumbs_up}" id="al-thumbs-widget-up"
        class="btn btn-md btn-info al-thumbs-widget" aria-label="{word('Thumbs up')}">{fa_icon('thumbs-up', size='md')}</a>
    <a href="{js_thumbs_down}" id="al-thumbs-widget-down"
        class="btn btn-md btn-info al-thumbs-widget" aria-label="{word('Thumbs down')}">{fa_icon('thumbs-down', size='md')}</a>
    """
    if review_action:
        review_area_id = review_action + "_area_id"
        js_review = (
            f"javascript:altoolbox_review_send('{review_action}', '{review_area_id}')"
        )
        widget += f"""
      <p class="al-review-text al-hidden">{word(review_display)}</p>
      <textarea class="datextarea al-review-text al-hidden" id="{review_area_id}"
          alt="{word('Write your review here')}" rows="4"></textarea>
      <br class="al-review-text">
      {action_button_html(js_review, label=word(submit_review_button), size='md', classname='al-review-text al-hidden')} 
        """
    widget += f"""
  <p class="al-after-review al-hidden">{word(post_review_display)}</p>
  </div>
</div>"""
    return widget


def sum_if_defined(*pargs):
    """Lets you add up the value of variables that are not in a list"""
    total = 0
    for source in pargs:
        if defined(source):
            total += value(source)
    return total


def add_records(obj, labels):
    """List demo interviews in the current package to be run from the landing page"""
    index = 0
    for key, val in labels.items():
        obj.appendObject()
        obj[index].name = key
        obj[index].description = val[0]
        obj[index].reference = val[1]
        index += 1
    return obj


def output_checkbox(
    value_to_check: bool, checked_value: str = "[X]", unchecked_value: str = "[  ]"
):
    """Generate a conditional checkbox for docx templates

    Args:
      checked_value: defaults to `[X]` but can be set to any string or even a `DAFile` or `DAStaticFile`
          with the image of a checkbox
      unchecked_value: opposite meaning of `checked_value` and defaults to `[  ]`

    """
    if value_to_check:
        return checked_value
    else:
        return unchecked_value


def nice_county_name(address: Address) -> str:
    """
    If the county name contains the word "County", which Google Address
    Autocomplete does by default, remove it.
    """
    if not hasattr(address, "county"):
        return ""
    if address.county.endswith(" County"):
        return address.county[: -len(" County")]
    else:
        return address.county



import re
from docassemble.base.util import CustomDataType, DAValidationError


class PhoneNumber(CustomDataType):
    name = "al_international_phone"
    input_class = "al_international_phone"
    javascript = """\
/** When combined with the phone number javascript file shown in
*    phone_number_validation_demo.yml, this docassemble CustomDataType
*    will make sure a user has give valid phone numbers in input fields
*    with the `datatype` `al_international_phone`. This includes international
*    numbers.
*
* This docassemble CustomDataType will be installed on your server along
*    with the ALToolbox package. Like all docassemble CustomDataTypes, it
*    will always be active on your server as long as it's there.
*    
* ## More Resources
*    1. What we're using: https://github.com/jackocnr/intl-tel-input
*    1. It uses https://www.npmjs.com/package/google-libphonenumber
*    1. https://github.com/google/libphonenumber/blob/master/FALSEHOODS.md
*    1. https://github.com/google/libphonenumber/blob/master/FAQ.md
*/
var validatePhoneNumber = function( value, element, params ) {
  /** Returns true if the international phone number is valid or if the field
  *    is empty. Otherwise returns false.
  */
  // When a field is empty, this value will be '', which counts as `false` here
  if ( value.trim() ) {
    // We can't use window.intlTelInputGlobals.loadUtils. It lets us
    // validate numbers docassemble has formatted, but only the first time
    // The user hits 'Back'. After that, the numbers are seen as invalid.
    
    // Get the special field that has already been created during page load
    var telLibObj = window.intlTelInputGlobals.getInstance( element );
    // Validate its value when the form is submitted
    return telLibObj.isValidNumber();
  }
  // If it's an empty field, then it's valid as far as this is concerned.
  return true;
};

$.validator.addMethod( 'al_international_phone', validatePhoneNumber );
"""
    jq_rule = "al_international_phone"
    # People that have just entered an invalid US phone number could find this confusing
    jq_message = 'This phone number doesn\'t look right. Note that a non-US number needs a "+" before the number.'

    # No server-side validation. Just avoiding user error here.
    # If you want to discuss that decision, make an issue on the repository.



# Based on https://github.com/GBLS/docassemble-income/blob/master/docassemble/income/income.py

from docassemble.base.util import (
    DAObject,
    DADict,
    DAList,
    DAOrderedDict,
    DAEmpty,
    Individual,
    comma_list,
    get_locale,
    word,
    log,
    object_name_convert,
    value,
)
from decimal import Decimal
import re
import datetime
import docassemble.base.functions
import json
from typing import Any, Dict, Callable, List, Optional, Set, Union, Tuple, Mapping

__all__ = [
    "times_per_year",
    "recent_years",
    "ALIncome",
    "ALIncomeList",
    "ALExpense",
    "ALExpenseList",
    "ALJob",
    "ALJobList",
    "ALAsset",
    "ALAssetList",
    "ALVehicle",
    "ALVehicleList",
    "ALSimpleValue",
    "ALSimpleValueList",
    "ALItemizedValue",
    "ALItemizedValueDict",
    "ALItemizedJob",
    "ALItemizedJobList",
]


def _currency_float_to_decimal(value: Union[str, float]) -> Decimal:
    """Given a float (that was set by a docassemble currency datatype, so
    rounded to the nearest `fractional_digit` decimal places), returns the
    exact decimal value, without floating point representation errors
    """
    if isinstance(value, float):
        # Print out the value of the float, rounded to the smallest allowable amount in the
        # locale currency, and use this value to make the exact Decimal value
        digits = get_locale("frac_digits")
        return Decimal(f"{value:.{digits}f}")
    else:
        return Decimal(value)


def times_per_year(
    times_per_year_list: List[Tuple[int, str]], times_per_year: float
) -> str:
    """
    Get the lower-case textual description that matches a time period contained
    in a "times per year" list.

    The goal of this function is to allow you to reflect the user's selection
    back to them, either on screen or in a document.

    In `al_income.yml` there is a default `times_per_year_list`, but the list
    that you use must be passed as a parameter as it's common to want to
    customize this for a given financial statement.

    For example: if the `times_per_year` is 12, it will return "monthly" from
    the default `times_per_year_list`.

    If the times per year does not exist in the given list, it will return a
    literal string like "five times per year".

    Fractional or floating point-based times_per_year are permissible in the
    times_per_year_list, although they are not commonly used. E.g., `.5` would
    represent "every two years". Items not contained in the list (to provide a
    specific lookup name) will have a string representation that is rounded to
    the nearest whole integer.
    """
    try:
        for row in times_per_year_list:
            if float(times_per_year) == float(row[0]):
                return row[1].lower()
        return (
            docassemble.base.functions.nice_number(int(times_per_year), capitalize=True)
            + " "
            + docassemble.base.functions.word("times per year")
        )
    except:
        return str(times_per_year)


def recent_years(
    past: int = 25, order: str = "descending", future: int = 1
) -> List[int]:
    """
    Returns a list of the most recent past years, continuing into the future.
    Defaults to most recent 15 years+1. Useful to populate a combobox of years
    where the most recent ones are most likely. E.g. automobile years or
    birthdate.

    Keyword parameters:
    * past {float} The number of past years to list, including the current year.
        The default is 15
    * order {string} 'descending' or 'ascending'. Default is `descending`.
    * future (defaults to 1).
    """
    now = datetime.datetime.now()
    if order == "ascending":
        return list(range(now.year - past, now.year + future, 1))
    else:
        return list(range(now.year + future, now.year - past, -1))


class ALPeriodicAmount(DAObject):
    """
    Represents an amount (could be an income or an expense depending on the context)
    that reoccurs some times per year. E.g, to express a weekly period, use 52. The default
    is 1 (a year).

    Attributes:
    .value {str | float | Decimal} A number representing an amount of money accumulated during
        the `times_per_year` of this income.
    .times_per_year {float | Decimal} Represents a number of the annual frequency of
        the income. E.g. 12 for a monthly income.
    .source {str} (Optional) The "source" of the income, like a "job" or a "house".
    .display_name {str} (Optional) If present, will have a translated string to show the
        user, as opposed to a raw english string from the program
    """

    def __str__(self) -> str:
        """Returns the income's `.total()` as string, not its object name."""
        return str(self.total())

    def total(self, times_per_year: float = 1) -> Decimal:
        """
        Returns the income over the specified times_per_year,

        To calculate `.total()`, an ALPeriodicAmount must have a `.times_per_year` and `.value`.
        """
        val = _currency_float_to_decimal(self.value)
        return (val * Decimal(self.times_per_year)) / Decimal(times_per_year)


class ALIncome(ALPeriodicAmount):
    """
    Represents an income which may have an hourly rate or a salary. Hourly rate
    incomes must include hours per period (times per year). Period is some
    denominator of a year. E.g, to express a weekly period, use 52. The default
    is 1 (a year).

    Attributes:
    .value {str | float | Decimal} A number representing an amount of money accumulated during
        the `times_per_year` of this income.
    .times_per_year {float | Decimal} Represents a number of the annual frequency of
        the income. E.g. 12 for a monthly income.
    .is_hourly {bool} (Optional) True if the income is hourly.
    .hours_per_period {float | Decimal} (Optional) If the income is hourly, the number of
        hours during the annual frequency of this job. E.g. if the annual
        frequency is 52 (weekly), the hours per week might be 50. That is, 50
        hours per week. This attribute is required if `.is_hourly` is True.
    .source {str} (Optional) The "source" of the income, like a "job" or a "house".
    .owner {str} (Optional) Full name of the income's owner as a single string.
    """

    def total(self, times_per_year: float = 1) -> Decimal:
        """
        Returns the income over the specified times_per_year, taking into account
        hours per period for hourly items. For example, for an hourly income of 10
        an hour, 40 hours a week, `income.total(1)` would be 20,800, the yearly income,
        and `income.total(52)` would be 400, the weekly income.

        To calculate `.total()`, an ALIncome must have a `.times_per_year` and `.value`.
        It can also have `.is_hourly` and `.hours_per_period`.
        """
        if hasattr(self, "is_hourly") and self.is_hourly:
            val = _currency_float_to_decimal(self.value)
            return (
                val * Decimal(self.hours_per_period) * Decimal(self.times_per_year)
            ) / Decimal(times_per_year)
        else:
            return super().total(times_per_year=times_per_year)


class ALExpense(ALPeriodicAmount):
    """Not much changes from ALPeriodic Amount, just the generic object questions"""

    pass


SourceType = Union[Set[str], List[str], str]


def _to_set(s: Optional[Union[Set, List, str]]) -> Set:
    """Converts a str, list of strings, or set of strings into a set of strings,
    which can be used to filter items in ALIncome classes.

    This is for internal use meant to ensure that `source` input is always a set.
    """
    if s is None:
        return set()
    if isinstance(s, set):
        return s
    if isinstance(s, list):
        return set(s)
    if isinstance(s, str):
        return set([s])
    return set()


def _source_to_callable(
    source: Optional[SourceType] = None, exclude_source: Optional[SourceType] = None
) -> Callable[[str], bool]:
    """Combines both a positive and negative lists into a single set that should be tested for inclusion"""
    exclude_set = _to_set(exclude_source)
    include_set = _to_set(source).difference(exclude_set)
    if include_set:
        return lambda s: s in include_set
    else:
        if exclude_set:
            return lambda s: s not in exclude_set
        else:
            return lambda s: True


class ALIncomeList(DAList):
    """
    Represents a filterable DAList of incomes-type items. It can make
    use of these attributes and methods in its items:

    .source
    .owner
    .times_per_year
    .value
    .total()
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        if not hasattr(self, "object_type") or self.object_type is None:
            self.object_type = ALIncome

    def sources(self) -> Set[str]:
        """Returns a set of the unique sources in the ALIncomeList."""
        sources = set()
        for item in self.elements:
            if hasattr(item, "source"):
                sources.add(item.source)
        return sources

    def matches(
        self, source: SourceType, exclude_source: Optional[SourceType] = None
    ) -> "ALIncomeList":
        """
        Returns an ALIncomeList consisting only of elements matching the specified
        income source, assisting in filling PDFs with predefined spaces. `source`
        may be a list.
        """
        # Always make sure we're working with a set
        satifies_sources = _source_to_callable(source, exclude_source)
        # Construct the filtered list
        return ALIncomeList(
            elements=[item for item in self.elements if satifies_sources(item.source)],
            object_type=self.object_type,
        )

    def total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
        owner: Optional[str] = None,
    ) -> Decimal:
        """
        Returns the total periodic value in the list, gathering the list items
        if necessary. You can optionally filter by `source`. `source` can be a
        string or a list. You can also filter by one `owner`.

        To calculate `.total()` correctly, all items must have a `.total()` and
        it should be a positive value. Job-type incomes should automatically
        exclude deductions.
        """
        self._trigger_gather()
        result: Decimal = Decimal(0)
        if times_per_year == 0:
            return result
        satisfies_sources = _source_to_callable(source, exclude_source)
        for item in self.elements:
            if (source is None and exclude_source is None) or (
                hasattr(item, "source") and satisfies_sources(item.source)
            ):
                if owner is None:  # if the user doesn't care who the owner is
                    result += Decimal(item.total(times_per_year=times_per_year))
                else:
                    if (
                        not (isinstance(owner, DAEmpty))
                        and hasattr(item, "owner")
                        and item.owner == owner
                    ):
                        result += Decimal(item.total(times_per_year=times_per_year))
        return result

    def move_checks_to_list(
        self,
        selected_types: Optional[DADict] = None,
        selected_terms: Optional[Mapping] = None,
    ):
        """Gives a 'gather by checklist' option.
        If no selected_types param is passed, requires that a .selected_types
        attribute be set by a `datatype: checkboxes` fields
        If "other" is in the selected_types, the source will not be set directly

        Sets the attribute "moved" to true, doesn't set gathered, because this isn't
        idempotent, so trying to also gather all info about the checks in the list doesn't
        work well.
        """
        if selected_types is None:
            selected_types = self.selected_types
        if not selected_terms:
            selected_terms = {}
        self.elements.clear()
        for source in selected_types.true_values(insertion_order=True):
            if source == "other":
                self.appendObject()
            else:
                self.appendObject(
                    source=source, display_name=selected_terms.get(source, source)
                )
        self.moved = True


class ALJob(ALIncome):
    """
    Represents a single job that may be hourly or pay-period based.

    The job can have a net and gross income figure, but it does not represent
    individual items like wages, tips or deductions that may appear on a paycheck--the
    user must enter the total amount for "net" and "gross" income for a given period.

    Can be stored in an ALJobList.

    Attributes:
    .value {float | Decimal} A number representing an amount of money accumulated during
        the `times_per_year` of this income.
    .times_per_year {float} Represents a number of the annual frequency of
        the value. E.g. 12 for a monthly value.
    .is_hourly {bool} (Optional): Whether the gross total should be calculated based on hours
        worked per week
    .hours_per_period {float} (Optional) The number of hours during the annual
        frequency of this job. E.g. if the annual frequency is 52 (weekly), the
        hours per week might be 50. That is, 50 hours per week.
    .deduction {float} (Optional) The amount of money deducted from the total value each period.
        If this job is hourly, deduction is still from each period, not each hour. Used to
        calculate the net income in `net_income()`.
    .employer {Individual} (Optional) A docassemble Individual object, employer.address is the address
        and employer.phone is the phone
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        # if not hasattr(self, "source") or self.source is None:
        #    self.source = "job"
        if not hasattr(self, "employer"):
            if hasattr(self, "employer_type"):
                self.initializeAttribute("employer", self.employer_type)
            else:
                self.initializeAttribute("employer", Individual)

    def gross_total(self, times_per_year: float = 1) -> Decimal:
        """
        Same as ALIncome total. Returns the income over the specified times_per_year,
        representing the `.value` attribute of the item.

        `times_per_year` is some denominator of a year. E.g. to express a weekly
        period, use 52. The default is 1 (a year).
        """
        return self.total(times_per_year=times_per_year)

    def deductions(self, times_per_year: float = 1) -> Decimal:
        """
        Returns the total deductions from someone's pay over the specificed times_per_year
        (not per hour if hourly).

        `times_per_year` is some denominator of a year. E.g. to express a weekly
        period, use 52. The default is 1 (a year).
        """
        deduction = _currency_float_to_decimal(self.deduction)
        return (deduction * Decimal(self.times_per_year)) / Decimal(times_per_year)

    def net_total(self, times_per_year: float = 1) -> Decimal:
        """
        Returns the net income over a time period, found using
        `self.value` and `self.deduction`.

        `times_per_year` is some denominator of a year. E.g, to express a weekly
        period, use 52. The default is 1 (a year).

        `self.deduction` is the amount deducted from one's pay over a period (not
        per hour if hourly).

        This will force the gathering of the ALJob's `.value` and `.deduction` attributes.
        """
        return self.total(times_per_year=times_per_year) - self.deductions(
            times_per_year=times_per_year
        )

    def employer_name_address_phone(self) -> str:
        """
        Returns name, address and phone number of employer as a string. Forces
        gathering the `.employer`, `.employer_address`, and `.employer_phone`
        attributes.
        """
        if self.employer.address.address and self.employer.phone:
            return (
                f"{self.employer.name}: {self.employer.address}, {self.employer.phone}"
            )
        if self.employer.address.address:
            return f"{self.employer.name}: {self.employer.address}"
        if self.employer.phone:
            return f"{self.employer.name}: {self.employer.phone}"
        return f"{self.employer.name}"

    def normalized_hours(self, times_per_year: float = 1) -> float:
        """
        Returns the normalized number of hours worked in a given times_per_year,
        based on the self.hours_per_period and self.times_per_year attributes.

        For example, if the person works 10 hours a week, it will return
        520 when the times_per_year parameter is 1.

        `times_per_year` is some denominator of a year. E.g, to express a weekly
        period, use 52. The default is 1 (a year).

        This will force the gathering of the attributes `.hours_per_period` and
        `.times_per_year`
        """
        return (float(self.hours_per_period) * int(self.times_per_year)) / float(
            times_per_year
        )


class ALJobList(ALIncomeList):
    """
    Represents a list of ALJobs. Adds the `.gross_total()` and
    `.net_total()` methods to the ALIncomeList class. It's a more common
    way of reporting income than ALItemizedJobList.
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.object_type = ALJob

    def total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
        owner: Optional[str] = None,
    ) -> Decimal:
        """
        Returns the sum of the gross incomes of its ALJobs divided by the time
        times_per_year. You can filter the jobs by `source`. `source` can be a
        string or a list.

        `times_per_year` is some denominator of a year. E.g, to express a weekly
        period, use 52. The default is 1 (a year).
        """
        return self.gross_total(
            times_per_year=times_per_year, source=source, exclude_source=exclude_source
        )

    def gross_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the sum of the gross incomes of its ALJobs divided by the time
        times_per_year. You can filter the jobs by `source`. `source` can be a
        string or a list.

        `times_per_year` is some denominator of a year. E.g, to express a weekly
        period, use 52. The default is 1 (a year).
        """
        self._trigger_gather()
        result: Decimal = Decimal(0)
        if times_per_year == 0:
            return result
        satisfies_sources = _source_to_callable(source, exclude_source)
        for job in self.elements:
            if satisfies_sources(job.source):
                result += Decimal(job.gross_total(times_per_year=times_per_year))
        return result

    def net_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the sum of the net incomes of its ALJobs divided by the time
        times_per_year. You can filter the jobs by `source`. `source` can be a
        string or a list. Leaving out `source` will use all sources.

        If the job is hourly, the `net_total()` may not be comparable to the
        `gross_total()`.

        `times_per_year` is some denominator of a year. E.g, to express a weekly
        period, use 52. The default is 1 (a year).
        """
        self._trigger_gather()
        result: Decimal = Decimal(0)
        if times_per_year == 0:
            return result
        satisfies_sources = _source_to_callable(source, exclude_source)
        for job in self.elements:
            if satisfies_sources(job.source):
                result += Decimal(job.net_total(times_per_year=times_per_year))
        return result

    def deductions(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the sum of the deductions of its ALJobs divided by the time
        times_per_year. You can filter the jobs by `source`. Leaving out `source`
        will use all sources.
        """
        self._trigger_gather()
        result: Decimal = Decimal(0)
        if times_per_year == 0:
            return result
        satisfies_sources = _source_to_callable(source, exclude_source)
        for job in self.elements:
            if satisfies_sources(job.source):
                result += Decimal(job.deductions(times_per_year=times_per_year))
        return result


class ALExpenseList(ALIncomeList):
    """
    A list of expenses

    * each element has a:
        * value
        * source
        * display name
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.object_type = ALExpense


class ALAsset(ALIncome):
    """
    An ALAsset represents an asset that has a market value, an optional income
    that the asset earns, and an optional balance which may be helpful if the
    asset represents a financial account rather than a physical asset.

    Can be stored in an ALAssetList.

    Attributes:
     .market_value {float | Decimal} Market value of the asset.
    .balance {float | Decimal } Current balance of the account, e.g., like
        the balance in a checking account, but could also represent a loan
        amount.
    .value {float | Decimal} (Optional) Represents the income the asset earns
        for a given `times_per_year`, such as interest earned in a checking
        account. If not defined, the income will be set to 0, to simplify
        representing the many common assets that do not earn any income.
    .times_per_year {float} (Optional) Number of times per year the asset
        earns the income listed in the `value` attribute.
    .owner {str} (Optional) Full name of the asset owner as a single string.
    .source {str} (Optional) The "source" of the asset, like "vase".
    """

    def total(self, times_per_year: float = 1) -> Decimal:
        """
        Returns the .value attribute divided by the times per year you want to
        calculate. The value defaults to 0.

        `times_per_year` is some denominator of a year. E.g, to express a weekly
        period, use 52. The default is 1 (a year).
        """
        if not hasattr(self, "value") or self.value == "":
            return Decimal(0)
        else:
            return super(ALAsset, self).total(times_per_year=times_per_year)


class ALAssetList(ALIncomeList):
    """
    A list of ALAssets. The `total()` of the list will be the total income
    earned, which may not be what you want for a list of assets. To get the
    total value of all assets, use the `market_value()` method.
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.object_type = ALAsset

    def market_value(
        self,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the total `.market_value` of assets in the list. You can filter
        the assets by `source`. `source` can be a string or a list.
        """
        result = Decimal(0)
        satisfies_sources = _source_to_callable(source, exclude_source)
        for asset in self.elements:
            if (source is None and exclude_source is None) or (
                satisfies_sources(asset.source)
            ):
                result += _currency_float_to_decimal(asset.market_value)
        return result

    def balance(
        self,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the total `.balance` of assets in the list,
        which typically corresponds to the available funds
        in a financial account.

        You can filter the assets by `source`. `source` can be a string or a list.
        """
        self._trigger_gather()
        result = Decimal(0)
        satisfies_sources = _source_to_callable(source, exclude_source)
        for asset in self.elements:
            if (source is None and exclude_source is None) or (
                satisfies_sources(asset.source)
            ):
                result += _currency_float_to_decimal(asset.balance)
        return result

    def owners(
        self,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Set[str]:
        """
        Returns a set of the unique owners of the assets.  You can filter the
        assets by `source`. `source` can be a string or a list.
        """
        owners = set()
        if source is None and exclude_source is None:
            for asset in self.elements:
                if hasattr(asset, "owner"):
                    owners.add(asset.owner)
        else:
            satisfies_source = _source_to_callable(source, exclude_source)
            for asset in self.elements:
                if (
                    hasattr(asset, "owner")
                    and hasattr(asset, "source")
                    and satisfies_source(asset.source)
                ):
                    owners.add(asset.owner)
        return owners


class ALVehicle(ALAsset):
    """
    An ALAsset with special attributes that help representing a vehicle.

    Vehicles have a .year_make_model() method which facilitates
    listing vehicles on many financial statement forms.

    Attributes:
    .year {str} The model year of the vehicle, like 2022
    .make {str} The make of the vehicle, like "Honda"
    .model {str} The model of the vehicle, like "Accord"
    .market_value {float | Decimal} Market value of an asset.
    .balance {float | Decimal} Balance of an asset.
    .value {float | Decimal} (Optional) Income earned by the vehicle (typically 0)
    .times_per_year {float} Time frequency over which the `value` is earned
    .owner {str} Full name of the asset owner as a single string.
    .source {str} (Optional) The "source" of the asset. Defaults to "vehicle".
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        if not hasattr(self, "source"):
            self.source = "vehicle"

    def year_make_model(self) -> str:
        """
        Returns a string of the format year/make/model of the vehicle. Triggers
        gathering those attributes.
        """
        return f"{ self.year } / { self.make } / { self.model }"


class ALVehicleList(ALAssetList):
    """List of ALVehicles. Extends ALAssetList."""

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.object_type = ALVehicle


class ALSimpleValue(DAObject):
    """
    Represents a currency value. It's meant to be stored in a list. Can be an
        item in an ALSimpleValueList.

    Attributes:
    .value {str | float } The monetary value of the item.
    .transaction_type {str} (Optional) Can be "expense", which will give a
        negative value to the total of the item.
    .source {str} (Optional) The "source" of the item, like "vase".
    """

    def total(self) -> Decimal:
        """
        If desired, to use as a ledger, values can be signed (mixed positive and
        negative). Setting transaction_type = 'expense' makes the value negative.
        Use min=0 in that case.

        If you use signed values, be careful when placing in an ALIncomeList
        object. The `total()` method may return unexpected results in that case.
        """
        val = _currency_float_to_decimal(self.value)
        if hasattr(self, "transaction_type"):
            return val * Decimal(-1) if (self.transaction_type == "expense") else val
        else:
            return val

    def __str__(self) -> str:
        """Returns the total as a formatted string"""
        return str(self.total())


class ALSimpleValueList(DAList):
    """Represents a filterable DAList of ALSimpleValues."""

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.object_type = ALSimpleValue

    def sources(self) -> Set:
        """
        Returns a set of the unique sources of values stored in the list.
        """
        sources = set()
        for value in self.elements:
            if hasattr(value, "source"):
                sources.add(value.source)
        return sources

    def total(
        self,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the total value in the list, gathering the list items if
        necessary. You can filter the values by `source`. `source` can be a
        string or a list.
        """
        self._trigger_gather()
        result = Decimal(0)
        if source is None and exclude_source is None:
            for value in self.elements:
                result += value.total()
        else:
            satisfies_source = _source_to_callable(source, exclude_source)
            for value in self.elements:
                if satisfies_source(value.source):
                    result += value.total()
        return result


class ALItemizedValue(DAObject):
    """
    An item in an ALItemizedValueDict (a line item like wages, tips or union dues).
    Should be a positive number, even if it will later be subtracted from the
    job's net total.

    WARNING: This item's period-based value can't be calculated correctly
    outside of an ALItemizedJob. Its value should only be accessed through the
    filtering methods of the ALItemizedJob that contains it.

    Attributes:
    .value {float | Decimal} A number representing an amount of money accumulated
        during the `times_per_year` of this item or this item's job.
    .is_hourly {bool} Whether this particular item is calculated hourly.
    .times_per_year { float} A denominator of a year representing the annual
         frequency of the job. E.g. 12 for monthly.
    .exists {bool} (Optional) Allows an interview author to pre-define some common
        descriptors, like "wages" or "union dues" without requiring the user to
        provide a value for each item.

        If the ".exists" attribute is False or undefined, the item will not be used
        when calculating totals.
    """

    def income_fields(self, use_exists=True) -> List[Dict[str, Any]]:
        """
        Returns a YAML structure representing the list of fields for an itemized value,
        to be passed to a `code` attribute of a question's fields
        """
        if use_exists:
            return [
                {
                    "label": self.display_name,
                    "field": self.attr_name("exists"),
                    "datatype": "yesno",
                },
                {
                    "label": word("Amount"),
                    "field": self.attr_name("value"),
                    "show if": self.attr_name("exists"),
                    "datatype": "currency",
                    "min": 0,
                },
            ]
        else:
            return [
                {
                    "label": self.display_name,
                    "field": self.attr_name("value"),
                    "datatype": "currency",
                    "min": 0,
                },
            ]

    def total(self) -> Decimal:
        # If an item's value doesn't exist, use a value of 0
        # TODO: is this behavior correct, or should it force gathering the value?
        # What does a no-value item in the list represent?
        if not hasattr(self, "value") or hasattr(self, "exists") and not self.exists:
            return Decimal(0)

        return _currency_float_to_decimal(self.value)

    def __str__(self) -> str:
        """Returns a string of the value of the item with two decimal places."""
        currency_str = "{:.2f}".format(self.value)
        return currency_str

    def __float__(self) -> float:
        if hasattr(self, "exists") and not self.exists:
            return 0.0
        else:
            return float(self.value)

    def __int__(self) -> int:
        return int(float(self))

    def __format__(self, format_spec) -> str:
        return f"{float(self):{format_spec}}"


class ALItemizedValueDict(DAOrderedDict):
    """
    Dictionary that can contain ALItemizedValues (e.g. line items) for an
    ALItemizedJob. E.g., wages, tips and deductions being the most common.

    An ALItemizedJob will have two ALItemizedValueDicts, one for income
    and one for deductions.

    WARNING: Should only be accessed through an ALItemizedJob. Otherwise
        you may get unexpected results.
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.object_type = ALItemizedValue
        if not hasattr(self, "complete_attribute"):
            self.complete_attribute = "value"

    def hook_after_gather(self) -> None:
        """
        Update item lists after they've been gathered or edited to remove non-existent
        items. Will still allow the developer to set `auto_gather=False` if they
        want without affecting this functionality.
        See https://docassemble.org/docs/objects.html#DAList.hook_after_gather.

        If a developer wants to remove these items _before_ gathering is finished,
        they can use similar code in their question's `validation code:`
        """
        keys_to_delete = []
        # During the loop, the list has to stay steady, so don't delete those items.
        # `.elements.items()` example of preventing gathering in `validation code:`
        for key, value in self.elements.items():
            if hasattr(value, "exists") and value.exists is False:
                keys_to_delete.append(key)
        # Delete the keys
        for key in keys_to_delete:
            self.delitem(key)

    def total(self) -> Decimal:
        val = Decimal(0)
        for key, value in self.elements.items():
            if hasattr(value, "exists") and not value.exists:
                continue
            val += _currency_float_to_decimal(value.value)
        return val

    def __str__(self) -> str:
        """
        Returns a string of the dictionary's key/value pairs as two-element lists in a list.
        E.g. '[["federal_taxes", "2500.00"], ["wages", "15.50"]]'
        """
        to_stringify = []
        for key in self:
            to_stringify.append((key, "{:.2f}".format(self[key].value)))
        pretty = json.dumps(to_stringify, indent=2)
        return pretty


class ALItemizedJob(DAObject):
    """
    An "Itemized" job is a job which allows the user to report very specific,
    granular details about the money that they earn in that job and any
    deductions that they have on their paycheck. This detailed accounting of
    money for each job is required on some financial statements, although in
    many financial statements, just reporting gross and net income is sufficient.

    For example, an ALItemizedJob can let the user report:
    - Wages at one hourly rate
    - Overtime at a second hourly rate
    - Tips earned during that time period
    - A fixed salary earned for that pay period
    - Union Dues
    - Insurance
    - Taxes

    If the financial statement only requires "gross" and "net" income, the
    ALJobList has a simpler API and may be the preferred way to represent the
    income in code.

    Attributes:
    .to_add {ALItemizedValueDict} Dict of ALItemizedValues that would be added
        to a job's net total, like wages and tips.
    .to_subtract {ALItemizedValueDict} Dict of ALItemizedValues that would be
        subtracted from a net total, like union dues or insurance premiums.
    .times_per_year {float} A denominator of a year, like 12 for monthly, that
        represents how frequently the income is earned
    .is_hourly {bool} (Optional) Whether the value represents a figure that the
        user earns on an hourly basis, rather than for the full time period
    .hours_per_period {int} (Optional) If the job is hourly, how many hours the
        user works per period.
    .employer {Individual} (Optional) Individual assumed to have a name and,
        optionally, an address and phone.
    .source {str} (Optional) The category of this item, like "public service".
        Defaults to "job".

    WARNING: Individual items in `.to_add` and `.to_subtract` should not be used
        directly. They should only be accessed through the filtering methods of
        this job.

    Fulfills these requirements:
    - A job can be hourly. Its wages will be calculated with that in mind.
    - Despite an hourly job, some individual items must be calculated using the
        job's whole period.
    - Some items will have their own periods.
    - In a list of jobs, a developer may need to access full time and part time
        jobs separately.
    - In a list of jobs, a developer may need to sum all items from one source,
        such as tips or taxes.
    - The developer needs access to total money coming in, total money going out,
        and the total of money going in and money coming out.
    - A user must be able to add their own arbitrary items.
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        # if not hasattr(self, "source") or self.source is None:
        #    self.source = "job"
        if not hasattr(self, "employer"):
            if hasattr(self, "employer_type"):
                self.initializeAttribute("employer", self.employer_type)
            else:
                self.initializeAttribute("employer", Individual)
        # Money coming in
        if not hasattr(self, "to_add"):
            self.initializeAttribute("to_add", ALItemizedValueDict)
        # Money being taken out
        if not hasattr(self, "to_subtract"):
            self.initializeAttribute("to_subtract", ALItemizedValueDict)

    def _item_value_per_times_per_year(
        self, item: ALItemizedValue, times_per_year: float = 1
    ) -> Decimal:
        """
        Given an ALItemizedValue and a times_per_year, returns the value
        accumulated by the item for that `times_per_year`, applying the
        attributes of the top-level ALItemizedJob, such as `times_per_year` and
        `is_hourly` as a default, and otherwise applying the attributes of the
        ALItemizedValue.

        `times_per_year` is some denominator of a year. E.g, to express a weekly
        period, use 52. The default is 1 (a year).

        Args:
        arg item {ALItemizedValue} Object containing the value and other props
            for an "in" or "out" ALItemizedJob item.
        kwarg: times_per_year {float} (Optional) Number of times per year you
            want to calculate. E.g, to express a weekly period, use 52. Default is 1.
        """
        if times_per_year == 0:
            return Decimal(0)

        # If an item has its own period, use that
        # Otherwise, default to the parent times_per_year
        if hasattr(item, "times_per_year") and item.times_per_year:
            frequency_to_use = item.times_per_year
        else:
            frequency_to_use = self.times_per_year

        # NOTE: fixes a bug that was present < 0.8.2
        try:
            hours_per_period = Decimal(self.hours_per_period)
        except:
            log(
                word(
                    "Your hours per period need to be just a single number, without words"
                ),
                "danger",
            )
            delattr(self, "hours_per_period")
            self.hours_per_period  # Will cause another exception

        # Both the job and the item itself need to be hourly to be
        # calculated as hourly
        is_hourly = self.is_hourly and hasattr(item, "is_hourly") and item.is_hourly
        value = item.total()

        # Use the appropriate calculation
        if is_hourly:
            return (
                value * Decimal(hours_per_period) * Decimal(frequency_to_use)
            ) / Decimal(times_per_year)
        else:
            return (value * Decimal(frequency_to_use)) / Decimal(times_per_year)

    def total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Alias for ALItemizedJob.gross_total to integrate with ALIncomeList math.
        """
        return self.gross_total(
            times_per_year=times_per_year, source=source, exclude_source=exclude_source
        )

    def gross_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the sum of positive values (payments) for a given times_per_year.
        You can filter the items by `source`. `source` can be a string or a list.
        If you use sources from deductions, they will be ignored.

        Args:
        kwarg: times_per_year {float} (Optional) Number of times per year you
            want to calculate. E.g, to express a weekly period, use 52. Default is 1.
        kwarg: source {str | [str]} (Optional) Source or list of sources of desired
            item(s).
        """
        # self.to_add._trigger_gather()
        total = Decimal(0)
        if times_per_year == 0:
            return total
        # Add up all money coming in from a source
        satisfies_sources = _source_to_callable(source, exclude_source)
        for key, value in self.to_add.elements.items():
            if satisfies_sources(key):
                total += self._item_value_per_times_per_year(
                    value, times_per_year=times_per_year
                )
        return total

    def deduction_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the sum of money going out (normally, deductions like union
        dues) divided by a pay times_per_year as a positive value. You can
        filter the items by `source`. `source` can be a string or a list.

        Args:
        kwarg: times_per_year {float} (Optional) Number of times per year you
            want to calculate. E.g, to express a weekly period, use 52. Default is 1.
        kwarg: source {str | List[str]} (Optional) Source or list of sources of desired
            item(s).
        """
        # self.to_subtract._trigger_gather()
        total = Decimal(0)
        if times_per_year == 0:
            return total
        # Make sure we're always working with a list of sources (names?)
        # Add up all money coming in from a source
        satisfies_sources = _source_to_callable(source, exclude_source)
        for key, value in self.to_subtract.elements.items():
            if satisfies_sources(key):
                total += self._item_value_per_times_per_year(
                    value, times_per_year=times_per_year
                )
        return total

    def net_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the net (gross minus deductions) value of the job divided by
        `times_per_year`. You can filter the items by `source`. `source` can be a
        string or a list. E.g. "full time" or ["full time", "union dues"]

        Args:
        kwarg: times_per_year {float} (Optional) Number of times per year you
            want to calculate. E.g, to express a weekly period, use 52. Default is 1.
        kwarg: source {str | List[str]} (Optional) Source or list of sources of desired
            item(s).
        """
        # self.to_add._trigger_gather()
        # self.to_subtract._trigger_gather()
        return self.gross_total(
            times_per_year=times_per_year, source=source, exclude_source=exclude_source
        ) - self.deduction_total(
            times_per_year=times_per_year, source=source, exclude_source=exclude_source
        )

    def employer_name_address_phone(self) -> str:
        """
        Returns concatenation of employer name and, if they exist, employer
        address and phone number.
        """
        info_list = []
        has_address = (
            hasattr(self.employer.address, "address") and self.employer.address.address
        )
        has_number = (
            hasattr(self.employer, "phone_number") and self.employer.phone_number
        )
        # Create a list so we can take advantage of `comma_list` instead
        # of doing further fiddly list manipulation
        if has_address:
            info_list.append(self.employer.address.on_one_line())
        if has_number:
            info_list.append(self.employer.phone_number)
        # If either exist, add a colon and the appropriate strings
        if has_address or has_number:
            return (
                f"{ self.employer.name.full(middle='full') }: {comma_list( info_list )}"
            )
        return self.employer.name.full(middle="full")

    def normalized_hours(self, times_per_year: float = 1) -> float:
        """
        Returns the normalized number of hours worked in a given times_per_year,
        based on the self.hours_per_period and self.times_per_year attributes.

        For example, if the person works 10 hours a week, it will return
        520 when the times_per_year parameter is 1.
        """
        return (float(self.hours_per_period) * float(self.times_per_year)) / float(
            times_per_year
        )


class ALItemizedJobList(DAList):
    """
    Represents a list of ALItemizedJobs that can have both payments and money
        out. This is a less common way of reporting income.
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        if not hasattr(self, "source") or self.source is None:
            self.source = "jobs"
        if not hasattr(self, "object_type") or self.object_type is None:
            self.object_type = ALItemizedJob

    def sources(self, which_side: Optional[str] = None) -> Set[str]:
        """Returns a set of the unique sources in all of the jobs.
        By default gets from both sides, if which_side is "deductions", only gets from deductions.
        """
        sources = set()
        if not which_side:
            which_side = "all"
        for job in self.elements:
            if which_side == "all" or which_side == "incomes":
                sources.update(job.to_add.keys())
            if which_side == "all" or which_side == "deductions":
                sources.update(job.to_subtract.keys())
        return sources

    def total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Alias for ALItemizedJobList.gross_total to integrate with
        ALIncomeList math.
        """
        return self.gross_total(
            times_per_year=times_per_year, source=source, exclude_source=exclude_source
        )

    def gross_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the sum of the gross incomes of the list's jobs divided by the
        times_per_year. You can filter the items by `source`. `source` can be a
        string or a list.

        Args:
        kwarg: source {str | [str]} - (Optional) Source or list of sources of
            desired job items to sum from every itemized job.
            E.g. ['tips', 'commissions']
        kwarg: times_per_year {float} (Optional) Number of times per year you
            want to calculate. E.g, to express a weekly period, use 52. Default is 1.
        """
        self._trigger_gather()
        total = Decimal(0)
        if times_per_year == 0:
            return total
        # Add all job gross totals from particular sources
        for job in self.elements:
            total += job.gross_total(
                times_per_year=times_per_year,
                source=source,
                exclude_source=exclude_source,
            )
        return total

    def deduction_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the sum of the deductions of the list's jobs divided by the
        times_per_year. You can filter the items by `source`. `source` can be a
        string or a list.

        Args:
        kwarg: source {str | [str]} - (Optional) Source or list of sources of
            desired job items to sum from every itemized job.
            E.g. ['taxes', 'dues']
        kwarg: times_per_year {float} (Optional) Number of times per year you
            want to calculate. E.g, to express a weekly period, use 52. Default is 1.
        """
        self._trigger_gather()
        total = Decimal(0)
        if times_per_year == 0:
            return total
        # Add all the money going out for all jobs
        for job in self.elements:
            total += job.deduction_total(
                times_per_year=times_per_year,
                source=source,
                exclude_source=exclude_source,
            )
        return total

    def net_total(
        self,
        times_per_year: float = 1,
        source: Optional[SourceType] = None,
        exclude_source: Optional[SourceType] = None,
    ) -> Decimal:
        """
        Returns the net of the list's jobs (money in minus money out) divided by
        the times_per_year. You can filter the items by `source`. `source` can be a
        string or a list.

        Args:
        kwarg: source {str | List[str]} - (Optional) Source or list of sources of
            desired job items to sum from every itemized job.
            E.g. ['tips', 'taxes']
        kwarg: times_per_year {float} (Optional) Number of times per year you
            want to calculate. E.g, to express a weekly period, use 52. Default is 1.
        """
        return self.gross_total(
            times_per_year=times_per_year, source=source
        ) - self.deduction_total(
            times_per_year=times_per_year, source=source, exclude_source=exclude_source
        )



import unittest

from decimal import Decimal
from .al_income import (
    ALIncome,
    ALIncomeList,
    ALAsset,
    ALAssetList,
    ALVehicle,
    ALVehicleList,
    ALSimpleValue,
    ALSimpleValueList,
    ALItemizedJob,
    ALItemizedJobList,
    ALItemizedValueDict,
    ALItemizedValue,
    recent_years,
)


class test_correct_outputs(unittest.TestCase):
    def test_simple_value(self):
        value = ALSimpleValue(transaction_type="expense", value=5)
        self.assertEqual(Decimal(-5), value.total())
        value.transaction_type = "asset"
        self.assertEqual(Decimal(5), value.total())

        value_str = ALSimpleValue(value="5.3")
        self.assertEqual(Decimal("5.3"), value_str.total())
        value_neg_str = ALSimpleValue(transaction_type="expense", value="5.3")
        self.assertEqual(Decimal("-5.3"), value_neg_str.total())
        value_float = ALSimpleValue(value=6.3)
        self.assertEqual(Decimal("6.3"), value_float.total())
        self.assertEqual("6.30", str(value_float.total()))

    def test_simple_value_list(self):
        val1 = ALSimpleValue(transaction_type="expense", source="real estate", value=5)
        val2 = ALSimpleValue(transaction_type="asset", source="job", value=10.1)
        val_list = ALSimpleValueList(elements=[val1, val2])
        self.assertEqual(Decimal("5.1"), val_list.total())
        self.assertSetEqual(set(["real estate", "job"]), val_list.sources())
        val_list.elements.append(
            ALSimpleValue(transaction_type="asset", source="job", value=20.1)
        )
        self.assertSetEqual(set(["real estate", "job"]), val_list.sources())

    def test_income(self):
        income = ALIncome(value=10.1, times_per_year=12)
        self.assertEqual(Decimal("121.2"), income.total())
        self.assertEqual(Decimal("10.1"), income.total(times_per_year=12))
        self.assertEqual(
            Decimal("2.33"), income.total(times_per_year=52).quantize(Decimal("0.01"))
        )

        hourly_income = ALIncome(
            value=4.4, times_per_year=52, is_hourly=True, hours_per_period=39
        )
        self.assertEqual(Decimal("8923.2"), hourly_income.total())
        self.assertEqual(Decimal("171.6"), hourly_income.total(52))

    def test_income_list(self):
        income = ALIncome(source="coding", value=12.53, times_per_year=12)
        hourly_income = ALIncome(
            source="coding",
            value=4.4,
            times_per_year=2,
            is_hourly=True,
            hours_per_period=41,
        )
        income_list = ALIncomeList(elements=[income, hourly_income])
        self.assertSetEqual(set(["coding"]), income_list.sources())

        self.assertEqual(Decimal(0), income_list.total(0))
        self.assertEqual(Decimal("511.16"), income_list.total(1))
        self.assertEqual(Decimal("511.16"), income_list.total(1, source="coding"))
        self.assertEqual(Decimal(0), income_list.total(1, source="wrong job"))
        self.assertEqual(
            Decimal("511.16"), income_list.total(1, source=["coding", "wrong job"])
        )

    def test_job(self):
        # TODO
        pass

    def test_job_list(self):
        # TODO
        pass

    def test_asset(self):
        home = ALAsset(market_value=1234567.89, source="home")
        self.assertEqual(Decimal(0), home.total())
        savings = ALAsset(
            balance=12.34, value=0.12, times_per_year=12, source="savings account"
        )
        self.assertEqual(Decimal("1.44"), savings.total())

    def test_asset_list(self):
        home = ALAsset(market_value=1234567.89, source="home")
        savings = ALAsset(
            balance=12.34, value=0.12, times_per_year=12, source="savings"
        )
        investing = ALAsset(
            balance=23.45, value=1.2, times_per_year=12, source="stocks"
        )
        checking = ALAsset(
            balance=34.56, value=0.01, times_per_year=12, source="checking"
        )
        asset_list = ALAssetList(elements=[home, savings, investing, checking])
        self.assertEqual(Decimal("1234567.89"), asset_list.market_value(source="home"))
        self.assertEqual(
            Decimal("35.79"), asset_list.balance(source=["savings", "stocks"])
        )
        self.assertEqual(
            Decimal("15.84"), asset_list.total(exclude_source=["checking"])
        )

    def test_vehicle(self):
        # TODO
        pass

    def test_vehicle_list(self):
        # TODO
        pass

    def test_itemized_value_dict(self):
        itemized_dict = ALItemizedValueDict(
            elements={
                "val1": ALItemizedValue(
                    value=5.30, exists=True, is_hourly=False, times_per_year=1
                ),
                "val2": ALItemizedValue(
                    value=5.30, exists=False, is_hourly=False, times_per_year=1
                ),
            }
        )
        itemized_dict.hook_after_gather()
        self.assertEqual(1, len(itemized_dict))

    def test_itemized_job(self):
        job = ALItemizedJob(
            is_hourly=True,
            is_part_time=True,
            name="Baby sitter",
            source="job",
            times_per_year=52,
        )
        job.to_add["part time"] = ALItemizedValue(is_hourly=True, value=10.04)
        job.hours_per_period = 10
        self.assertEqual(Decimal("5220.80"), job.gross_total())
        job.to_add["tips"] = ALItemizedValue(is_hourly=False, value=200.23)
        job.to_subtract["snacks"] = ALItemizedValue(is_hourly=False, value="24.21")
        self.assertEqual(Decimal("15632.76"), job.gross_total())
        self.assertEqual("15632.76", str(job.gross_total()))
        self.assertEqual(Decimal("1258.92"), job.deduction_total())
        self.assertEqual(Decimal("104.91"), job.deduction_total(times_per_year=12))
        self.assertEqual(Decimal("14373.84"), job.net_total())
        self.assertEqual("14373.84", str(job.net_total()))

    def test_itemized_job_list(self):
        job = ALItemizedJob(
            is_hourly=True,
            is_part_time=True,
            name="Baby sitter",
            source="job",
            times_per_year=52,
            hours_per_period=10,
        )
        job.to_add["part time"] = ALItemizedValue(is_hourly=True, value=10.04)
        job.to_add["tips"] = ALItemizedValue(is_hourly=False, value=200.23)
        job.to_subtract["snacks"] = ALItemizedValue(is_hourly=False, value="24.21")
        job_list = ALItemizedJobList(elements=[job])
        self.assertEqual(Decimal("15632.76"), job_list.gross_total())
        self.assertEqual(Decimal("1258.92"), job_list.deduction_total())
        self.assertEqual(Decimal("104.91"), job_list.deduction_total(times_per_year=12))
        self.assertEqual(Decimal("14373.84"), job_list.net_total())


if __name__ == "__main__":
    unittest.main()



import unittest
from .business_days import (
    is_business_day,
    get_next_business_day,
    get_date_after_n_business_days,
)
from docassemble.base.util import today, as_datetime


class TestBusinessDays(unittest.TestCase):
    def test_is_business_day(self):
        self.assertFalse(is_business_day("2022-09-05"))
        self.assertTrue(is_business_day("2022-09-06"))
        this_years_christmas = today().replace(month=12, day=25)
        self.assertFalse(is_business_day(this_years_christmas))

    def test_get_date_after_n_business_days(self):
        self.assertEqual(
            get_date_after_n_business_days("2022-12-24", 5), as_datetime("2023-01-03")
        )
        self.assertEqual(
            get_date_after_n_business_days("2022-12-12", 3), as_datetime("2022-12-15")
        )
        self.assertEqual(
            get_date_after_n_business_days(
                "2022-12-12",
                3,
                add_holidays={"12-13": "fake day 1", "12-14": "fake day2"},
            ),
            as_datetime("2022-12-19"),
        )
        self.assertEqual(
            get_date_after_n_business_days(
                "2022-12-24", 5, remove_holidays=["Christmas Day"]
            ),
            as_datetime("2022-12-30"),
        )


if __name__ == "__main__":
    unittest.main()



import holidays
import pandas as pd
import datetime
from datetime import date as dt
from docassemble.base.util import as_datetime, DADateTime
from typing import Union, Dict, Iterable, Mapping, Optional

"""
  External docs: 
  1. https://github.com/dr-prodigy/python-holidays (holidays module v0.13, as of 2/2022)
  2. https://python-holidays.readthedocs.io/en/latest/examples.html  
  3. https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html  
"""

__all__ = [
    "standard_holidays",
    "non_business_days",
    "is_business_day",
    "get_next_business_day",
    "get_date_after_n_business_days",
]


def standard_holidays(
    year,
    country="US",
    subdiv="MA",
    add_holidays: Optional[Mapping] = None,
    remove_holidays: Optional[Iterable[str]] = None,
) -> holidays.HolidayBase:
    """
    Get all holidays in the specified year, country, and state (or other subdivision).
    Note that this draws on the "holidays" package which may deviate slightly from
    holidays observed by a local court, but should be very close to accurate.

    add_holidays should be a dictionary from dates ("12-15") to the name of the holiday.

    Returns a dictionary like-object that you can treat like:
    {
        "2021-01-01": "New Year's Day",
        ...
        "2021-12-25": "Christmas Day",
    }

    In place of a string, the object that is returned can also be treated as though
    the keys are datetime.date objects.
    """
    # 1. Get standard holidays from python's holidays module
    countr_holidays: holidays.HolidayBase = holidays.country_holidays(
        country=country, subdiv=subdiv, years=year
    )

    # 2. Remove known obsolete holidays
    if country == "US" and subdiv == "MA":
        countr_holidays.pop_named("Evacuation Day")

    # 3. Remove user specified holidays
    if remove_holidays:
        for record in remove_holidays:
            countr_holidays.pop_named(record)

    # 4. Append user defined holidays
    if add_holidays:
        for k, v in add_holidays.items():
            # Attach the given year to the key
            key = f"{year}-{k}"
            countr_holidays[key] = v
    return countr_holidays


def non_business_days(
    year,
    country="US",
    subdiv="MA",
    add_holidays: Optional[Mapping] = None,
    remove_holidays: Optional[Iterable[str]] = None,
    first_n_dates=0,
    last_n_dates=0,
) -> dict:
    # TODO: this function may not be necessary, but check with @purplesky2016 before removing
    # 1. Collect weekends and standard holidays
    # 1.1 Get all saturdays and sundays in the given year

    sundays = (
        pd.date_range(start=str(year), end=str(year + 1), freq="W-SUN")
        .strftime("%Y-%m-%d")
        .tolist()
    )
    saturdays = (
        pd.date_range(start=str(year), end=str(year + 1), freq="W-SAT")
        .strftime("%Y-%m-%d")
        .tolist()
    )

    # 1.2 Get holidays of the given country and subdivision (state/province) in the year
    local_holidays = standard_holidays(
        year=year,
        country=country,
        subdiv=subdiv,
        add_holidays=add_holidays,
        remove_holidays=remove_holidays,
    )

    # 2. Populate date_dict using as_datetime format as key for later sorting
    # 2.1 Populate date_dict with holidays
    date_dict = {}
    for raw_date, name in local_holidays.items():
        date_dict[as_datetime(raw_date)] = name

    # 2.2 Add weekends if not already in date_dict
    for a in saturdays:
        if as_datetime(a) not in date_dict.keys():
            date_dict[as_datetime(a)] = "Saturday"
    for b in sundays:
        if as_datetime(b) not in date_dict.keys():
            date_dict[as_datetime(b)] = "Sunday"

    # 3. Sort date_dict then change key from as_datetime to mm/dd/yyyy for easier application
    date_dict = dict(sorted(date_dict.items()))

    clean_date_dict = {}
    for k, v in date_dict.items():
        clean_date_dict[k.strftime("%Y-%m-%d")] = v

    # 4. Take a subset if user desires it (useful only if this function is explicitly called)
    final_output = {}
    if first_n_dates > 0 and last_n_dates > 0:
        first_slice = {
            s: clean_date_dict[s] for s in list(clean_date_dict)[:first_n_dates]
        }
        last_slice = {
            s: clean_date_dict[s]
            for s in list(clean_date_dict)[len(clean_date_dict.keys()) - last_n_dates :]
        }
        final_output = {**first_slice, **last_slice}  # Add two dicts
    elif first_n_dates > 0:
        final_output = {
            s: clean_date_dict[s] for s in list(clean_date_dict)[:first_n_dates]
        }
    elif last_n_dates > 0:
        final_output = {
            s: clean_date_dict[s]
            for s in list(clean_date_dict)[len(clean_date_dict.keys()) - last_n_dates :]
        }
    else:
        final_output = clean_date_dict

    # 5. Return the result
    return final_output


def is_business_day(
    date: Union[str, DADateTime],
    country="US",
    subdiv="MA",
    add_holidays: Optional[Mapping] = None,
    remove_holidays: Optional[Iterable[str]] = None,
) -> bool:
    """
    Returns true if and only if the specified date is a business day (i.e., not a holiday)
    in the specified jurisdiction. Business days are considered to be:
    1. weekdays other than Saturday and Sunday and
    1. days that are not a federal or state-observed holiday

    Example:
    ```
    assert(is_business_day("2023-03-26") == False)
    ```
    """
    if not isinstance(date, DADateTime):
        date = as_datetime(date)
    if date.dow in [
        6,
        7,
    ]:  # Docassemble codes Saturday and Sunday as 6 and 7 respectively
        return False
    if date.format("yyyy-MM-dd") in standard_holidays(
        year=date.year,
        country=country,
        subdiv=subdiv,
        add_holidays=add_holidays,
        remove_holidays=remove_holidays,
    ):
        return False
    return True


def get_next_business_day(
    start_date: Union[str, DADateTime],
    wait_n_days=1,
    country="US",
    subdiv="MA",
    add_holidays: Optional[Mapping] = None,
    remove_holidays: Optional[Iterable[str]] = None,
) -> DADateTime:
    """
    Returns the first day AFTER the specified start date that is
    not a federal or state holiday, Saturday or Sunday. Optionally,
    specify the parameter `wait_n_days` to get the first business day after
    at least, e.g., 10 days.

    Relies on the Python holidays package, which has fairly good support for
    holidays around the world and in various states and provinces, but local
    court rules may differ. You can see what holidays are used at
    https://github.com/dr-prodigy/python-holidays/tree/master/holidays/countries

    Args:
      start_date: the date to start with. Can be a date-formatted string (i.e. "2023-03-37", or
          "3-27-2023") or a DADateTime object
      wait_n_days: the number of days to find the find the date after. If 0, it returns the given
          date if it's a business day.
      country: the county to use business days from
      subdiv: the subdivision (e.g. state or province) to use business days from
      add_holidays: a dictionary from the date string ("12/25") to the name of the holiday,
          will add those holidays to be considered
      remove_holidays: the list of date strings ("12/25") of dates that are no longer holidays
    """
    if not isinstance(start_date, DADateTime):
        start_date = as_datetime(start_date)
    date_to_check = start_date.plus(days=wait_n_days)

    while not is_business_day(
        date_to_check,
        country=country,
        subdiv=subdiv,
        add_holidays=add_holidays,
        remove_holidays=remove_holidays,
    ):
        date_to_check = date_to_check.plus(days=1)
    return date_to_check


def get_date_after_n_business_days(
    start_date: Union[str, DADateTime],
    wait_n_days=1,
    country="US",
    subdiv="MA",
    add_holidays: Optional[Mapping] = None,
    remove_holidays: Optional[Iterable[str]] = None,
) -> DADateTime:
    """
    Returns a time period which contains a minimum of `n` business days.

    Args:
      start_date: the date to start with. Can be a date-formatted string (i.e. "2023-03-37", or
          "3-27-2023") or a DADateTime object
      wait_n_days: the number of businesses days to wait for. For example, `start_date` is a
          Friday, and `wait_n_days` is 2, then the date returned will be the next Tuesday.
      country: the county to use business days from
      subdiv: the subdivision (e.g. state or province) to use business days from
      add_holidays: a dictionary from the date string ("12/25") to the name of the holiday,
          will add those holidays to be considered
      remove_holidays: the list of date strings ("12/25") of dates that are no longer holidays
    """
    if not isinstance(start_date, DADateTime):
        start_date = as_datetime(start_date)
    date_to_check = start_date

    for _ in range(wait_n_days):
        date_to_check = date_to_check.plus(days=1)
        while not is_business_day(
            date_to_check,
            country=country,
            subdiv=subdiv,
            add_holidays=add_holidays,
            remove_holidays=remove_holidays,
        ):
            date_to_check = date_to_check.plus(days=1)
    return date_to_check



import re
from .copy_button import *
from base64 import b64encode


def display_template(
    template,
    scrollable=True,
    collapse=False,
    copy=False,
    classname="bg-light",
    class_name=None,  # deprecated
) -> str:
    # 1. Initialize
    if scrollable:
        scroll_class = "scrollable-panel"
        adjust_height = ""
    else:
        scroll_class = "not-scrollable"
        adjust_height = (
            f"onmouseover=\"this.style.height = (this.scrollHeight) + 'px';\""
        )

    # Introducing `classname` to try to align with `collapse_template`
    if not class_name:
        class_name = classname
    class_name = class_name.strip()

    container_classname = "al_display_template"

    container_id = (
        b64encode(str(template.instanceName).encode()).decode().replace("=", "")
    )
    contents_id = f"{ container_id }_contents"

    subject_html = ""
    if not template.subject == "":
        subject_html = f'<div class="panel-heading"><h3 class="subject">{template.subject_as_html(trim=True)}</h3></div>'

    # 2. If copiable, call copy_button_html() to generate the template content along with a copy button
    if copy:
        contents = copy_button_html(
            template,
            copy_template_block=True,
            scroll_class=scroll_class,
            style_class=class_name,
            adjust_height=adjust_height,
        )

        # 2.1 If collapsible, add collapsible elements to the output
        if collapse:
            return f'<div id="{container_id}" class="{container_classname}"><a class="collapsed" data-bs-toggle="collapse" href="#{contents_id}" role="button" aria-expanded="false" aria-controls="collapseExample"><span class="toggle-icon pdcaretopen"><i class="fas fa-caret-down"></i></span><span class="toggle-icon pdcaretclosed"><i class="fas fa-caret-right"></i></span><span class="subject">{template.subject_as_html(trim=True)}</span></a><div class="collapse" id="{contents_id}">{contents}</div></div>'

        # 2.2 If not collapsible, simply return output from copy_button_html()
        else:
            return f"""
<div id="{container_id}" class="{container_classname}">
{subject_html}
{contents}
</div>
"""

    # 3. If not copiable, generate the whole output
    else:
        if collapse:
            return f'<div id="{container_id}" class="{container_classname}"><a class="collapsed" data-bs-toggle="collapse" href="#{contents_id}" role="button" aria-expanded="false" aria-controls="collapseExample"><span class="toggle-icon pdcaretopen"><i class="fas fa-caret-down"></i></span><span class="toggle-icon pdcaretclosed"><i class="fas fa-caret-right"></i></span><span class="subject">{template.subject_as_html(trim=True)}</span></a><div class="collapse" id="{contents_id}"><div class="{scroll_class} card card-body {class_name} pb-1">{template.content_as_html()}</div></div></div>'

        else:
            return f'<div id="{container_id}" class="{container_classname} {scroll_class} card card-body {class_name} pb-1" id="{contents_id}">{subject_html}<div>{template.content_as_html()}</div></div>'



import os
import sys
from setuptools import setup, find_packages
from fnmatch import fnmatchcase
from distutils.util import convert_path

standard_exclude = ('*.pyc', '*~', '.*', '*.bak', '*.swp*')
standard_exclude_directories = ('.*', 'CVS', '_darcs', './build', './dist', 'EGG-INFO', '*.egg-info')

def find_package_data(where='.', package='', exclude=standard_exclude, exclude_directories=standard_exclude_directories):
    out = {}
    stack = [(convert_path(where), '', package)]
    while stack:
        where, prefix, package = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if os.path.isdir(fn):
                bad_name = False
                for pattern in exclude_directories:
                    if (fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()):
                        bad_name = True
                        break
                if bad_name:
                    continue
                if os.path.isfile(os.path.join(fn, '__init__.py')):
                    if not package:
                        new_package = name
                    else:
                        new_package = package + '.' + name
                        stack.append((fn, '', new_package))
                else:
                    stack.append((fn, prefix + name + '/', package))
            else:
                bad_name = False
                for pattern in exclude:
                    if (fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()):
                        bad_name = True
                        break
                if bad_name:
                    continue
                out.setdefault(package, []).append(prefix+name)
    return out

setup(name='docassemble.ALToolbox',
      version='0.8.3',
      description=('Collection of small utility functions, classes, and web components for Docassemble interviews'),
      long_description='# ALToolbox\r\n\r\n[![PyPI version](https://badge.fury.io/py/docassemble-ALToolbox.svg)](https://badge.fury.io/py/docassemble-ALToolbox)\r\n\r\nThis repository is used to host small Python modules, widgets, and JavaScript web components js files that enhance Docassemble interviews. These modules were\r\nbuilt as part of the Suffolk University Law School LIT Lab\'s [Document Assembly Line project](https://suffolklitlab.org/docassemble-AssemblyLine-documentation/).\r\nThey are placed here\r\nrather than in https://github.com/SuffolkLitLab/docassemble-AssemblyLine because we believe these small components can easily be used\r\nby anyone, regardless of whether they use any other code from the Document Assembly Line project.\r\n\r\nIf you want to add a small function to this project, consider adding it to the existing misc.py to avoid creating too many module files.\r\n\r\n## Documentation\r\n\r\nRead the [documentation for the functions and components](https://suffolklitlab.org/docassemble-AssemblyLine-documentation/docs/framework/altoolbox) to learn\r\nhow to use these components in your own [Docassemble](https://github.com/jhpyle/docassemble) projects.\r\n\r\n## Suffolk LIT Lab Document Assembly Line\r\n\r\n<img src="https://user-images.githubusercontent.com/7645641/142245862-c2eb02ab-3090-4e97-9653-bb700bf4c54d.png" alt="drawing of people working together on a website UI" width="300" style="align: center;"/>\r\n\r\nThe Assembly Line Project is a collection of volunteers, students, and institutions who joined together\r\nduring the COVID-19 pandemic to help increase access to the court system. Our vision is mobile-friendly,\r\neasy to use **guided** online forms that help empower litigants to access the court remotely.\r\n\r\nOur signature project is [CourtFormsOnline.org](https://courtformsonline.org).\r\n\r\nWe designed a step-by-step, assembly line style process for automating court forms on top of Docassemble\r\nand built several tools along the way that **you** can use in your home jurisdiction.\r\n\r\nThis package contains **runtime code** and **pre-written questions** to support authoring robust, \r\nconsistent, and attractive Docassemble interviews that help complete court forms.\r\n\r\nRead more on our [documentation page](https://suffolklitlab.org/docassemble-AssemblyLine-documentation/).\r\n\r\n\r\n## Related repositories\r\n\r\n* https://github.com/SuffolkLitLab/docassemble-AssemblyLine\r\n* https://github.com/SuffolkLitLab/docassemble-ALWeaver\r\n* https://github.com/SuffolkLitLab/docassemble-ALMassachusetts\r\n* https://github.com/SuffolkLitLab/docassemble-MassAccess\r\n* https://github.com/SuffolkLitLab/docassemble-ALThemeTemplate\r\n* https://github.com/SuffolkLitLab/EfileProxyServer\r\n\r\n## Contributors:\r\n* @plocket  \r\n* @nonprofittechy\r\n* @purplesky2016\r\n* @brycestevenwilley\r\n',
      long_description_content_type='text/markdown',
      author='AssemblyLine',
      author_email='52798256+plocket@users.noreply.github.com',
      license='The MIT License (MIT)',
      url='https://suffolklitlab.org/docassemble-AssemblyLine-documentation/docs/framework/altoolbox',
      packages=find_packages(),
      namespace_packages=['docassemble'],
      install_requires=['holidays>=0.27.1', 'pandas>=1.5.3'],
      zip_safe=False,
      package_data=find_package_data(where='docassemble/ALToolbox/', package='docassemble.ALToolbox'),
     )

